# -*- coding: utf-8 -*-
"""
Module to perform bias correction of renewable energy resource data
"""
import numpy as np
import logging

from rex.utilities.bc_utils import QuantileDeltaMapping

logger = logging.getLogger(__name__)


def _irrad_pre_proc(ghi, dni, dhi):
    """Irradiance data pre processing to get ancillary variables
    (run before bias correction).

    Parameters
    ----------
    ghi : np.ndarray
        2D array of global horizontal irradiance values in shape (time, space)
    dni : np.ndarray
        2D array of direct normal irradiance values in shape (time, space)
    dhi : np.ndarray
        2D array of diffuse horizontal irradiance values in shape (time, space)

    Returns
    -------
    ghi_zeros : np.ndarray
        2D boolean array, True where ghi==0, same shape as ghi input
    dni_zeros : np.ndarray
        2D boolean array, True where dni==0, same shape as dni input
    dhi_zeros : np.ndarray
        2D boolean array, True where dhi==0, same shape as dhi input
    cos_sza : np.ndarray
        2D array for cos(solar_zenith_angle) calculated from the basic
        relationship ``cos_sza = (ghi - dhi) / dni``
    """

    ghi_zeros = ghi == 0
    dni_zeros = dni == 0
    dhi_zeros = dhi == 0
    with np.errstate(divide='ignore', invalid='ignore'):
        cos_sza = (ghi - dhi) / dni

    return ghi_zeros, dni_zeros, dhi_zeros, cos_sza


def _irrad_post_proc(ghi, dni, ghi_zeros, dni_zeros, dhi_zeros, cos_sza):
    """Irradiance data post processing to calculate DHI and set limits on
    irradiance variables (run after bias correction).

    Parameters
    ----------
    ghi : np.ndarray
        2D array of global horizontal irradiance values in shape (time, space)
    dni : np.ndarray
        2D array of direct normal irradiance values in shape (time, space)
    ghi_zeros : np.ndarray
        2D boolean array, True where ghi==0, same shape as ghi input
    dni_zeros : np.ndarray
        2D boolean array, True where dni==0, same shape as dni input
    dhi_zeros : np.ndarray
        2D boolean array, True where dhi==0, same shape as dhi input
    cos_sza : np.ndarray
        2D array for cos(solar_zenith_angle) calculated from the basic
        relationship ``cos_sza = (ghi - dhi) / dni``

    Returns
    -------
    ghi : np.ndarray
        2D array of global horizontal irradiance values in shape (time, space)
    dni : np.ndarray
        2D array of direct normal irradiance values in shape (time, space)
    dhi : np.ndarray
        2D array of diffuse horizontal irradiance values in shape (time, space)
    """

    ghi = np.maximum(0, ghi)
    dni = np.maximum(0, dni)

    ghi[ghi_zeros] = 0
    dni[dni_zeros] = 0

    with np.errstate(divide='ignore', invalid='ignore'):
        dhi = ghi - (dni * cos_sza)

    dhi[dni_zeros] = ghi[dni_zeros]
    dhi[dhi_zeros] = 0
    dhi = np.maximum(0, dhi)

    assert not np.isnan(dhi).any()

    return ghi, dni, dhi


def lin_irrad(ghi, dni, dhi, scalar=1, adder=0):
    """Correct GHI and DNI using linear correction factors. Both irradiance
    variables are corrected as ``irradiance * scalar + adder``. DHI is
    preserved based on the relationship ``dhi = ghi - (dni * cos(sza))``. Times
    when GHI and DNI are zero are preserved and negative values are protected
    against.

    Parameters
    ----------
    ghi : np.ndarray
        2D array of global horizontal irradiance values in shape (time, space)
    dni : np.ndarray
        2D array of direct normal irradiance values in shape (time, space)
    dhi : np.ndarray
        2D array of diffuse horizontal irradiance values in shape (time, space)
    scalar : np.ndarray
        1D array of linear scalar values in the shape (space,)
    adder : np.ndarray
        1D array of linear adder values in the shape (space,)

    Returns
    -------
    ghi : np.ndarray
        2D array of global horizontal irradiance values in shape (time, space)
    dni : np.ndarray
        2D array of direct normal irradiance values in shape (time, space)
    dhi : np.ndarray
        2D array of diffuse horizontal irradiance values in shape (time, space)
    """

    ghi_zeros, dni_zeros, dhi_zeros, cos_sza = _irrad_pre_proc(ghi, dni, dhi)

    ghi = ghi * scalar + adder
    dni = dni * scalar + adder

    ghi, dni, dhi = _irrad_post_proc(ghi, dni, ghi_zeros, dni_zeros, dhi_zeros,
                                     cos_sza)

    return ghi, dni, dhi


def lin_ws(ws, scalar=1, adder=0):
    """Correct windspeed using linear correction factors. Windspeed is
    corrected as ``windspeed * scalar + adder`` with a minimum of zero.

    Parameters
    ----------
    ws : np.ndarray
        2D array of windspeed values in shape (time, space)
    scalar : np.ndarray
        1D array of linear scalar values in the shape (space,)
    adder : np.ndarray
        1D array of linear adder values in the shape (space,)

    Returns
    -------
    ws : np.ndarray
        2D array of windspeed values in shape (time, space)
    """
    ws = ws * scalar + adder
    ws = np.maximum(ws, 0)
    return ws


def qdm_irrad(ghi, dni, dhi,
              ghi_params_oh, dni_params_oh,
              ghi_params_mh, dni_params_mh,
              ghi_params_mf=None, dni_params_mf=None,
              dist='empirical', relative=True,
              sampling='linear', log_base=10):
    """Correct irradiance using the quantile delta mapping based on the method
    from Cannon et al., 2015

    Cannon, A. J., Sobie, S. R. & Murdock, T. Q. Bias Correction of GCM
    Precipitation by Quantile Mapping: How Well Do Methods Preserve Changes in
    Quantiles and Extremes? Journal of Climate 28, 6938–6959 (2015).

    Parameters
    ----------
    ghi : np.ndarray
        2D array of global horizontal irradiance values in shape (time, space)
    dni : np.ndarray
        2D array of direct normal irradiance values in shape (time, space)
    dhi : np.ndarray
        2D array of diffuse horizontal irradiance values in shape (time, space)
    ghi_params_oh : np.ndarray | list
        2D array of **observed historical** distribution parameters created
        from a multi-year set of data where the shape is (space, N). This
        can be the output of a parametric distribution fit like
        ``scipy.stats.weibull_min.fit()`` where N is the number of
        parameters for that distribution, or this can define the x-values
        of N points from an empirical CDF that will be linearly
        interpolated between. If this is an empirical CDF, this must
        include the 0th and 100th percentile values and have even
        percentile spacing between values.
    dni_params_oh : np.ndarray | list
        Same requirements as ghi_params_oh. This input arg is for the
        **observed historical distribution** for DNI.
    ghi_params_mh : np.ndarray | list
        Same requirements as ghi_params_oh. This input arg is for the **modeled
        historical distribution** for GHI.
    dni_params_mh : np.ndarray | list
        Same requirements as ghi_params_oh. This input arg is for the **modeled
        historical distribution** for DNI.
    ghi_params_mf : np.ndarray | list | None
        Same requirements as ghi_params_oh. This input arg is for the **modeled
        future distribution** for GHI. If this is None, this defaults to
        ghi_params_mh (no future data, just corrected to modeled historical
        distribution)
    dni_params_mf : np.ndarray | list | None
        Same requirements as ghi_params_oh. This input arg is for the **modeled
        future distribution** for DNI. If this is None, this defaults to
        dni_params_mh. (no future data, just corrected to modeled historical
        distribution)
    dist : str | np.ndarray
        Probability distribution name to use to model the data which
        determines how the param args are used. This can "empirical" or any
        continuous distribution name from ``scipy.stats``. Can also be a 1D
        array of dist inputs if being used from reV, but they must all be
        the same option.
    relative : bool | np.ndarray
        Flag to preserve relative rather than absolute changes in
        quantiles. relative=False (default) will multiply by the change in
        quantiles while relative=True will add. See Equations 4-6 from
        Cannon et al., 2015 for more details. Can also be a 1D array of
        dist inputs if being used from reV, but they must all be the same
        option.
    sampling : str | np.ndarray
        If dist="empirical", this is an option for how the quantiles were
        sampled to produce the params inputs, e.g., how to sample the
        y-axis of the distribution (see sampling functions in
        ``rex.utilities.bc_utils``). "linear" will do even spacing, "log"
        will concentrate samples near quantile=0, and "invlog" will
        concentrate samples near quantile=1. Can also be a 1D array of dist
        inputs if being used from reV, but they must all be the same option.
    log_base : int | float | np.ndarray
        Log base value if sampling is "log" or "invlog". A higher value
        will concentrate more samples at the extreme sides of the
        distribution. Can also be a 1D array of dist inputs if being used from
        reV, but they must all be the same option.

    Returns
    -------
    ghi : np.ndarray
        2D array of global horizontal irradiance values in shape (time, space)
    dni : np.ndarray
        2D array of direct normal irradiance values in shape (time, space)
    dhi : np.ndarray
        2D array of diffuse horizontal irradiance values in shape (time, space)
    """

    ghi_zeros, dni_zeros, dhi_zeros, cos_sza = _irrad_pre_proc(ghi, dni, dhi)

    ghi_qdm = QuantileDeltaMapping(ghi_params_oh, ghi_params_mh,
                                   ghi_params_mf, dist=dist,
                                   relative=relative, sampling=sampling,
                                   log_base=log_base)
    dni_qdm = QuantileDeltaMapping(dni_params_oh, dni_params_mh,
                                   dni_params_mf, dist=dist,
                                   relative=relative, sampling=sampling,
                                   log_base=log_base)

    # This will prevent inverse CDF functions from returning zero resulting in
    # a divide by zero error in the calculation of the QDM delta. These zeros
    # get fixed later in _irrad_post_proc
    ghi[ghi_zeros] = 1
    dni[dni_zeros] = 1

    ghi = ghi_qdm(ghi)
    dni = dni_qdm(dni)

    ghi, dni, dhi = _irrad_post_proc(ghi, dni, ghi_zeros, dni_zeros, dhi_zeros,
                                     cos_sza)

    return ghi, dni, dhi


def qdm_ws(ws, params_oh, params_mh, params_mf=None, dist='empirical',
           relative=True, sampling='linear', log_base=10):
    """Correct windspeed using quantile delta mapping based on the method from
    Cannon et al., 2015

    Cannon, A. J., Sobie, S. R. & Murdock, T. Q. Bias Correction of GCM
    Precipitation by Quantile Mapping: How Well Do Methods Preserve Changes in
    Quantiles and Extremes? Journal of Climate 28, 6938–6959 (2015).

    Parameters
    ----------
    ws : np.ndarray
        2D array of windspeed values in shape (time, space)
    params_oh : np.ndarray | list
        2D array of **observed historical** distribution parameters created
        from a multi-year set of data where the shape is (space, N). This
        can be the output of a parametric distribution fit like
        ``scipy.stats.weibull_min.fit()`` where N is the number of
        parameters for that distribution, or this can define the x-values
        of N points from an empirical CDF that will be linearly
        interpolated between. If this is an empirical CDF, this must
        include the 0th and 100th percentile values and have even
        percentile spacing between values.
    params_mh : np.ndarray | list
        Same requirements as params_oh. This input arg is for the **modeled
        historical distribution**.
    params_mf : np.ndarray | list | None
        Same requirements as params_oh. This input arg is for the **modeled
        future distribution**. If this is None, this defaults to params_mh
        (no future data, just corrected to modeled historical distribution)
    dist : str | np.ndarray
        Probability distribution name to use to model the data which
        determines how the param args are used. This can "empirical" or any
        continuous distribution name from ``scipy.stats``. Can also be a 1D
        array of dist inputs if being used from reV, but they must all be
        the same option.
    relative : bool | np.ndarray
        Flag to preserve relative rather than absolute changes in
        quantiles. relative=False (default) will multiply by the change in
        quantiles while relative=True will add. See Equations 4-6 from
        Cannon et al., 2015 for more details. Can also be a 1D array of
        dist inputs if being used from reV, but they must all be the same
        option.
    sampling : str | np.ndarray
        If dist="empirical", this is an option for how the quantiles were
        sampled to produce the params inputs, e.g., how to sample the
        y-axis of the distribution (see sampling functions in
        ``rex.utilities.bc_utils``). "linear" will do even spacing, "log"
        will concentrate samples near quantile=0, and "invlog" will
        concentrate samples near quantile=1. Can also be a 1D array of dist
        inputs if being used from reV, but they must all be the same option.
    log_base : int | float | np.ndarray
        Log base value if sampling is "log" or "invlog". A higher value
        will concentrate more samples at the extreme sides of the
        distribution. Can also be a 1D array of dist inputs if being used from
        reV, but they must all be the same option.

    Returns
    -------
    ws : np.ndarray
        2D array of windspeed values in shape (time, space)
    """

    qdm = QuantileDeltaMapping(params_oh, params_mh, params_mf, dist=dist,
                               relative=relative, sampling=sampling,
                               log_base=log_base)

    # This will prevent inverse CDF functions from returning zero resulting in
    # a divide by zero error in the calculation of the QDM delta. These zeros
    # get fixed later
    ws_zeros = ws == 0
    ws[ws_zeros] = 0.01

    ws = qdm(ws)

    ws = np.maximum(ws, 0)
    ws[ws_zeros] = 0

    return ws



# -*- coding: utf-8 -*-
"""
ResourceX Command Line Interface
"""
import click
import os
import logging
import pandas as pd

from rex.resource_extraction.resource_extraction import (ResourceX,
                                                         MultiFileResourceX)
from rex.utilities.loggers import init_logger
from rex.utilities.utilities import check_res_file
from rex import __version__

logger = logging.getLogger(__name__)


def _parse_sites(sites):
    """
    Parse sites

    Parameters
    ----------
    sites : str
        Path to .csv of .json containing sites to extract
        (sites is synonymous with gids aka spatial indices)

    Returns
    -------
    name : str
        sites file name
    gid : list | None
        Gids to extract
    lat_lon : list | None
        Lat, lon pairs to extract
    """
    name = os.path.splitext(os.path.basename(sites))[0]
    if sites.endswith('.csv'):
        sites = pd.read_csv(sites)
    elif sites.endswith('.json'):
        sites = pd.read_json(sites)
    else:
        raise RuntimeError("'--sites' must be a .csv or .json file!")

    if 'gid' in sites:
        gid = sites['gid'].values
        lat_lon = None
    elif 'latitude' in sites and 'longitude' in sites:
        gid = None
        lat_lon = sites[['latitude', 'longitude']].values
    else:
        raise RuntimeError('Must supply site "gid"s or "latitude" and '
                           '"longitude" as columns in "--sites" file')

    return name, gid, lat_lon


@click.group()
@click.version_option(version=__version__)
@click.option('--resource_h5', '-h5', required=True,
              type=click.Path(),
              help=('Path to Resource .h5 file'))
@click.option('--out_dir', '-o', required=True, type=click.Path(),
              help='Directory to dump output files')
@click.option('--log_file', '-log', default=None, type=click.Path(),
              show_default=True,
              help='Path to .log file, if None only log to stdout')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, resource_h5, out_dir, log_file, verbose):
    """
    ResourceX Command Line Interface
    """
    ctx.ensure_object(dict)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    ctx.obj['H5'] = resource_h5
    ctx.obj['OUT_DIR'] = out_dir
    ctx.obj['CLS_KWARGS'] = {}

    multi_h5_res, hsds = check_res_file(resource_h5)
    if multi_h5_res and not hsds:
        assert os.path.exists(os.path.dirname(resource_h5))
        ctx.obj['CLS'] = MultiFileResourceX
    elif hsds:
        ctx.obj['CLS_KWARGS']['hsds'] = hsds
    else:
        assert os.path.exists(resource_h5)

        ctx.obj['CLS'] = ResourceX

    if verbose:
        log_level = 'DEBUG'
    else:
        log_level = 'INFO'

    if log_file is not None:
        log_dir = os.path.dirname(log_file)
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)

    init_logger('rex', log_file=log_file, log_level=log_level)

    logger.info('Extracting Resource data from {}'.format(resource_h5))
    logger.info('Outputs to be stored in: {}'.format(out_dir))


@main.command()
@click.option('--lat_lon', '-ll', nargs=2, type=float,
              default=None, show_default=True,
              help='(lat, lon) coordinates of interest')
@click.option('--gid', '-gid', type=int, default=None, show_default=True,
              help='Resource gid of interest')
@click.option('--sites', '-s', type=click.Path(exists=True), default=None,
              show_default=True,
              help=('.csv or .json file with columns "latitude", "longitude" '
                    'OR "gid"'))
@click.pass_context
def sam_datasets(ctx, lat_lon, gid, sites):
    """
    Extract all datasets needed for SAM for the nearest pixel(s) to the given
    (lat, lon) coordinates, the given resource gid, or the give sites
    """
    if not lat_lon:
        lat_lon = None

    inputs = set((lat_lon, gid, sites))
    if len(inputs) == 1:
        msg = "Must supply '--lat-lon', '--gid', or '--sites'!"
        raise click.Abort(msg)
    elif len(inputs) > 2:
        msg = ("You must only supply one of '--lat-lon', '--gid', or "
               "'--sites'!")
        raise click.Abort(msg)

    if lat_lon or gid:
        logger.info('Saving data to {}'.format(ctx.obj['OUT_DIR']))
        with ctx.obj['CLS'](ctx.obj['H5'], **ctx.obj['CLS_KWARGS']) as f:
            if lat_lon is not None:
                f.get_SAM_lat_lon(lat_lon, out_path=ctx.obj['OUT_DIR'])
            elif gid is not None:
                f.get_SAM_gid(gid, out_path=ctx.obj['OUT_DIR'])

    else:
        name, gid, lat_lon = _parse_sites(sites)
        with ctx.obj['CLS'](ctx.obj['H5'], **ctx.obj['CLS_KWARGS']) as f:
            meta = f['meta']
            if lat_lon is not None:
                SAM_df = f.get_SAM_lat_lon(lat_lon)
            elif gid is not None:
                SAM_df = f.get_SAM_gid(gid)

        gids = []
        for df in SAM_df:
            gids.append(int(df.name.split('-')[-1]))
            out_path = "{}-{}.csv".format(df.name, name)
            out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
            logger.info('Saving data to {}'.format(out_path))
            df.to_csv(out_path)

        out_path = "{}-meta.csv".format(name)
        out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
        meta = meta.loc[gids]
        logger.info('Saving meta data to {}'.format(out_path))
        meta.to_csv(out_path)


@main.group()
@click.option('--dataset', '-d', type=str, required=True,
              help='Dataset to extract')
@click.pass_context
def dataset(ctx, dataset):
    """
    Extract a single dataset
    """
    ctx.obj['DATASET'] = dataset


@dataset.command()
@click.option('--lat_lon', '-ll', nargs=2, type=float,
              default=None, show_default=True,
              help='(lat, lon) coordinates of interest')
@click.option('--gid', '-gid', type=int, default=None, show_default=True,
              help='Resource gid of interest')
@click.pass_context
def site(ctx, lat_lon, gid):
    """
    Extract the nearest pixel to the given (lat, lon) coordinates OR the given
    resource gid
    """
    if lat_lon and gid is None:
        msg = "Must supply '--lat-lon' OR '--gid'!"
        raise click.Abort(msg)
    elif lat_lon and gid:
        msg = "You must only supply '--lat-lon' OR '--gid'!"
        raise click.Abort(msg)

    dataset = ctx.obj['DATASET']
    with ctx.obj['CLS'](ctx.obj['H5'], **ctx.obj['CLS_KWARGS']) as f:
        if lat_lon:
            site_df = f.get_lat_lon_df(dataset, lat_lon)
            gid = site_df.name
        elif gid is not None:
            site_df = f.get_gid_df(dataset, gid)

    out_path = "{}-{}.csv".format(dataset, gid)
    out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
    logger.info('Saving data to {}'.format(out_path))
    site_df.to_csv(out_path)


@dataset.command()
@click.option('--region', '-r', type=str, required=True,
              help='Region to extract')
@click.option('--region_col', '-col', type=str, default='state',
              show_default=True,
              help='Meta column to search for region')
@click.option('--timestep', '-ts', type=str, default=None, show_default=True,
              help='Timestep to extract')
@click.pass_context
def region(ctx, region, region_col, timestep):
    """
    Extract all pixels in the given region
    """
    dataset = ctx.obj['DATASET']
    if timestep is None:
        with ctx.obj['CLS'](ctx.obj['H5'], **ctx.obj['CLS_KWARGS']) as f:
            region_df = f.get_region_df(dataset, region, region_col=region_col)
            meta = f['meta']

        out_path = "{}-{}.csv".format(dataset, region)
        out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
        logger.info('Saving data to {}'.format(out_path))
        region_df.to_csv(out_path)

        out_path = "{}-meta.csv".format(region)
        out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
        meta = meta.loc[region_df.columns]
        logger.info('Saving meta data to {}'.format(out_path))
        meta.to_csv(out_path)
    else:
        with ctx.obj['CLS'](ctx.obj['H5'], **ctx.obj['CLS_KWARGS']) as f:
            map_df = f.get_timestep_map(dataset, timestep, region=region,
                                        region_col=region_col)

        out_path = "{}-{}-{}.csv".format(dataset, region, timestep)
        out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
        logger.info('Saving data to {}'.format(out_path))
        map_df.to_csv(out_path)


@dataset.command()
@click.option('--lat_lon_1', '-ll1', nargs=2, type=float,
              required=True,
              help='One corner of the bounding box')
@click.option('--lat_lon_2', '-ll2', nargs=2, type=float,
              required=True,
              help='The other corner of the bounding box')
@click.option('--file_suffix', '-fs', default=None, show_default=True,
              help='File name suffix')
@click.option('--timestep', '-ts', type=str, default=None, show_default=True,
              help='Timestep to extract')
@click.pass_context
def box(ctx, lat_lon_1, lat_lon_2, file_suffix, timestep):
    """
    Extract all pixels in the given bounding box
    """
    dataset = ctx.obj['DATASET']
    if file_suffix is None:
        file_suffix = 'box'

    if timestep is None:
        with ctx.obj['CLS'](ctx.obj['H5'], **ctx.obj['CLS_KWARGS']) as f:
            region_df = f.get_box_df(dataset, lat_lon_1, lat_lon_2)
            meta = f['meta']

        out_path = "{}-{}.csv".format(dataset, file_suffix)
        out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
        logger.info('Saving data to {}'.format(out_path))
        region_df.to_csv(out_path)

        out_path = "{}-meta.csv".format(file_suffix)
        out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
        meta = meta.loc[region_df.columns]
        logger.info('Saving meta data to {}'.format(out_path))
        meta.to_csv(out_path)
    else:
        with ctx.obj['CLS'](ctx.obj['H5'], **ctx.obj['CLS_KWARGS']) as f:
            map_df = f.get_timestep_map(dataset, timestep,
                                        box=(lat_lon_1, lat_lon_2))

        out_path = "{}-{}-{}.csv".format(dataset, file_suffix, timestep)
        out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
        logger.info('Saving data to {}'.format(out_path))
        map_df.to_csv(out_path)


@dataset.command()
@click.option('--sites', '-s', type=click.Path(exists=True), required=True,
              help=('.csv or .json file with columns "latitude", "longitude" '
                    'OR "gid"'))
@click.pass_context
def multi_site(ctx, sites):
    """
    Extract multiple sites given in '--sites' .csv or .json as
    "latitude", "longitude" pairs OR "gid"s
    """
    name, gid, lat_lon = _parse_sites(sites)
    dataset = ctx.obj['DATASET']
    with ctx.obj['CLS'](ctx.obj['H5'], **ctx.obj['CLS_KWARGS']) as f:
        meta = f['meta']
        if lat_lon is not None:
            site_df = f.get_lat_lon_df(dataset, lat_lon)
        elif gid is not None:
            site_df = f.get_gid_df(dataset, gid)

    out_path = "{}-{}.csv".format(dataset, name)
    out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
    logger.info('Saving data to {}'.format(out_path))
    site_df.to_csv(out_path)

    out_path = "{}-meta.csv".format(name)
    out_path = os.path.join(ctx.obj['OUT_DIR'], out_path)
    if not os.path.exists(out_path):
        meta = meta.loc[site_df.columns]
        logger.info('Saving meta data to {}'.format(out_path))
        meta.to_csv(out_path)
    else:
        logger.debug('Meta data has already been saved: {}'.format(out_path))


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running ResourceX CLI')
        raise



# -*- coding: utf-8 -*-
"""
Classes to handle h5 output files.
"""
import json
import logging
import numpy as np
import pandas as pd
import time
import sys
import click
import h5py
import h5pyd
import scipy

from rex.version import __version__
from rex.utilities.exceptions import (HandlerRuntimeError, HandlerValueError,
                                      ResourceKeyError)
from rex.resource import BaseResource
from rex.utilities.parse_keys import parse_keys, parse_slice
from rex.utilities.utilities import to_records_array

logger = logging.getLogger(__name__)


class Outputs(BaseResource):
    """
    Base class to handle output data in .h5 format

    Examples
    --------
    The Outputs handler can be used to initialize h5 files in the standard
    reV/rex resource data format.

    >>> from rex import Outputs
    >>> import pandas as pd
    >>> import numpy as np
    >>>
    >>> meta = pd.DataFrame({'latitude': np.ones(100),
    >>>                      'longitude': np.ones(100)})
    >>>
    >>> time_index = pd.date_range('20210101', '20220101', freq='1h',
    >>>                            closed='right')
    >>>
    >>> with Outputs('test.h5', 'w') as f:
    >>>     f.meta = meta
    >>>     f.time_index = time_index

    You can also use the Outputs handler to read output h5 files from disk.
    The Outputs handler will automatically parse the meta data and time index
    into the expected pandas objects (DataFrame and DatetimeIndex,
    respectively).

    >>> with Outputs('test.h5') as f:
    >>>     print(f.meta.head())
    >>>
         latitude  longitude
    gid
    0         1.0        1.0
    1         1.0        1.0
    2         1.0        1.0
    3         1.0        1.0
    4         1.0        1.0

    >>> with Outputs('test.h5') as f:
    >>>     print(f.time_index)
    DatetimeIndex(['2021-01-01 01:00:00+00:00', '2021-01-01 02:00:00+00:00',
                   '2021-01-01 03:00:00+00:00', '2021-01-01 04:00:00+00:00',
                   '2021-01-01 05:00:00+00:00', '2021-01-01 06:00:00+00:00',
                   '2021-01-01 07:00:00+00:00', '2021-01-01 08:00:00+00:00',
                   '2021-01-01 09:00:00+00:00', '2021-01-01 10:00:00+00:00',
                   ...
                   '2021-12-31 15:00:00+00:00', '2021-12-31 16:00:00+00:00',
                   '2021-12-31 17:00:00+00:00', '2021-12-31 18:00:00+00:00',
                   '2021-12-31 19:00:00+00:00', '2021-12-31 20:00:00+00:00',
                   '2021-12-31 21:00:00+00:00', '2021-12-31 22:00:00+00:00',
                   '2021-12-31 23:00:00+00:00', '2022-01-01 00:00:00+00:00'],
                  dtype='datetime64[ns, UTC]', length=8760, freq=None)

    There are a few ways to use the Outputs handler to write data to a file.
    Here is one example using the pre-initialized file we created earlier.
    Note that the Outputs handler will automatically scale float data using
    the "scale_factor" attribute. The Outputs handler will unscale the data
    while being read unless the unscale kwarg is explicityly set to False.
    This behavior is intended to reduce disk storage requirements for big
    data and can be disabled by setting dtype=np.float32 or dtype=np.float64
    when writing data.

    >>> Outputs.add_dataset(h5_file='test.h5', dset_name='dset1',
    >>>                     dset_data=np.ones((8760, 100)) * 42.42,
    >>>                     attrs={'scale_factor': 100}, dtype=np.int32)


    >>> with Outputs('test.h5') as f:
    >>>     print(f['dset1'])
    >>>     print(f['dset1'].dtype)
    [[42.42 42.42 42.42 ... 42.42 42.42 42.42]
     [42.42 42.42 42.42 ... 42.42 42.42 42.42]
     [42.42 42.42 42.42 ... 42.42 42.42 42.42]
     ...
     [42.42 42.42 42.42 ... 42.42 42.42 42.42]
     [42.42 42.42 42.42 ... 42.42 42.42 42.42]
     [42.42 42.42 42.42 ... 42.42 42.42 42.42]]
    float32

    >>> with Outputs('test.h5', unscale=False) as f:
    >>>     print(f['dset1'])
    >>>     print(f['dset1'].dtype)
    [[4242 4242 4242 ... 4242 4242 4242]
     [4242 4242 4242 ... 4242 4242 4242]
     [4242 4242 4242 ... 4242 4242 4242]
     ...
     [4242 4242 4242 ... 4242 4242 4242]
     [4242 4242 4242 ... 4242 4242 4242]
     [4242 4242 4242 ... 4242 4242 4242]]
    int32

    Note that the Outputs handler is specifically designed to read and
    write spatiotemporal data. It is therefore important to intialize the meta
    data and time index objects even if your data is only spatial or only
    temporal. Furthermore, the Outputs handler will always assume that 1D
    datasets represent scalar data (non-timeseries) that corresponds to the
    meta data shape, and that 2D datasets represent spatiotemporal data whose
    shape corresponds to (len(time_index), len(meta)). You can see these
    constraints here:

    >>> Outputs.add_dataset(h5_file='test.h5', dset_name='bad_shape',
                            dset_data=np.ones((1, 100)) * 42.42,
                            attrs={'scale_factor': 100}, dtype=np.int32)
    HandlerValueError: 2D data with shape (1, 100) is not of the proper
    spatiotemporal shape: (8760, 100)

    >>> Outputs.add_dataset(h5_file='test.h5', dset_name='bad_shape',
                            dset_data=np.ones((8760,)) * 42.42,
                            attrs={'scale_factor': 100}, dtype=np.int32)
    HandlerValueError: 1D data with shape (8760,) is not of the proper
    spatial shape: (100,)
    """

    def __init__(self, h5_file, mode='r', unscale=True, str_decode=True,
                 group=None):
        """
        Parameters
        ----------
        h5_file : str
            Path to .h5 resource file
        mode : str, optional
            Mode to instantiate h5py.File instance, by default 'r'
        unscale : bool, optional
            Boolean flag to automatically unscale variables on extraction,
            by default True
        str_decode : bool, optional
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read,
            by default True
        group : str, optional
            Group within .h5 resource file to open, by default None
        """
        super().__init__(h5_file, unscale=unscale, hsds=False,
                         str_decode=str_decode, group=group, mode=mode)
        self._mode = mode
        self._group = self._check_group(group)
        self._shape = None

        if self.writable:
            self.set_version_attr()

    def __len__(self):
        _len = 0
        if 'meta' in self.datasets:
            _len = self.h5['meta'].shape[0]

        return _len

    def __setitem__(self, keys, arr):
        if self.writable:
            ds, ds_slice = parse_keys(keys)

            slice_test = False
            if isinstance(ds_slice, tuple):
                slice_test = ds_slice[0] == slice(None, None, None)

            if ds.endswith('meta') and slice_test:
                self._set_meta(ds, arr)
            elif ds.endswith('time_index') and slice_test:
                self._set_time_index(ds, arr)
            else:
                self._set_ds_array(ds, arr, ds_slice)

    @property
    def full_version_record(self):
        """Get record of versions for dependencies

        Returns
        -------
        dict
            Dictionary of package versions for dependencies
        """
        versions = {'rex': __version__,
                    'pandas': pd.__version__,
                    'numpy': np.__version__,
                    'python': sys.version,
                    'click': click.__version__,
                    'h5py': h5py.__version__,
                    'h5pyd': h5pyd.__version__,
                    'scipy': scipy.__version__
                    }
        return versions

    def set_version_attr(self):
        """Set the version attribute to the h5 file."""
        self.h5.attrs['version'] = __version__
        self.h5.attrs['full_version_record'] = json.dumps(
            self.full_version_record)
        self.h5.attrs['package'] = 'rex'

    @property
    def version(self):
        """
        Version of package used to create file

        Returns
        -------
        str
        """
        return self.h5.attrs['version']

    @property
    def package(self):
        """
        Package used to create file

        Returns
        -------
        str
        """
        return self.h5.attrs['package']

    @property
    def source(self):
        """
        Package and version used to create file

        Returns
        -------
        str
        """
        out = ("{}_{}"
               .format(self.h5.attrs['package'], self.h5.attrs['version']))
        return out

    @property
    def shape(self):
        """
        Variable array shape from time_index and meta

        Returns
        -------
        tuple
            shape of variables arrays == (time, locations)
        """
        if self._shape is None:
            dsets = self.datasets
            if 'meta' in dsets:
                self._shape = self.h5['meta'].shape
                if 'time_index' in dsets:
                    self._shape = self.h5['time_index'].shape + self._shape

        return self._shape

    @property
    def writable(self):
        """
        Check to see if h5py.File instance is writable

        Returns
        -------
        is_writable : bool
            Flag if mode is writable
        """
        is_writable = True
        mode = ['a', 'w', 'w-', 'x']
        if self._mode not in mode:
            is_writable = False

        return is_writable

    @BaseResource.meta.setter  # pylint: disable-msg=E1101
    def meta(self, meta):
        """
        Write meta data to disk, convert type if neccessary

        Parameters
        ----------
        meta : pandas.DataFrame | numpy.recarray
            Locational meta data
        """
        self._set_meta('meta', meta)

    @BaseResource.time_index.setter  # pylint: disable-msg=E1101
    def time_index(self, time_index):
        """
        Write time_index to dics, convert type if neccessary

        Parameters
        ----------
        time_index : pandas.DatetimeIndex | ndarray
            Temporal index of timesteps
        """
        self._set_time_index('time_index', time_index)

    @property
    def SAM_configs(self):
        """
        SAM configuration JSONs used to create CF profiles

        Returns
        -------
        configs : dict
            Dictionary of SAM configuration JSONs
        """
        if 'meta' in self.datasets:
            configs = {k: json.loads(v)
                       for k, v in self.h5['meta'].attrs.items()}
        else:
            configs = {}

        return configs

    @property
    def run_attrs(self):
        """
        Runtime attributes stored at the global (file) level

        Returns
        -------
        global_attrs : dict
        """
        return self.global_attrs

    @run_attrs.setter
    def run_attrs(self, run_attrs):
        """
        Set runtime attributes as global (file) attributes

        Parameters
        ----------
        run_attrs : dict
            Dictionary of runtime attributes (args, kwargs)
        """
        if self.writable:
            for k, v in run_attrs.items():
                self.h5.attrs[k] = v

    @staticmethod
    def _check_data_dtype(dset_name, data, dtype, attrs=None):
        """
        Check data dtype and scale if needed

        Parameters
        ----------
        dset_name : str
            Name of dataset being written to disk
        data : ndarray
            Data to be written to disc
        dtype : str
            dtype of data on disc
        attrs : dict, optional
            Attributes to be set. May include 'scale_factor',
            by default None

        Returns
        -------
        data : ndarray
            Data ready for writing to disc:
            - Scaled and converted to dtype
        """
        if attrs is None:
            attrs = {}

        scale_factor = attrs.get('scale_factor', None)

        scale = (scale_factor is not None
                 and not np.issubdtype(data.dtype, np.integer))
        if scale:
            if scale_factor != 1 and not np.issubdtype(dtype, np.integer):
                msg = ('Output dtype for "{}" must be an integer in '
                       'order to apply scale factor {}".'
                       .format(dset_name, scale_factor))
                logger.error(msg)
                raise HandlerRuntimeError(msg)

            data_type_differs = not np.issubdtype(data.dtype, np.dtype(dtype))
            is_integer = np.issubdtype(dtype, np.integer)
            if data_type_differs and is_integer:
                # apply scale factor and dtype
                data = np.round(data * scale_factor).astype(dtype)

        elif (not np.issubdtype(data.dtype, np.dtype(dtype))
                and not np.issubdtype(np.dtype(dtype), np.floating)):
            msg = ('A scale_factor is needed to scale '
                   '"{}" of type "{}" to "{}".'
                   .format(dset_name, data.dtype, dtype))
            raise HandlerRuntimeError(msg)

        return data

    def _check_group(self, group):
        """
        Ensure group is in .h5 file

        Parameters
        ----------
        group : str
            Group of interest
        """
        if group is not None:
            if group not in self._h5:
                try:
                    if self.writable:
                        self._h5.create_group(group)
                except Exception as ex:
                    msg = ('Cannot create group {}: {}'
                           .format(group, ex))
                    raise HandlerRuntimeError(msg) from ex

        return group

    def _set_meta(self, ds, meta, attrs=None):
        """
        Write meta data to disk

        Parameters
        ----------
        ds : str
            meta dataset name
        meta : pandas.DataFrame | numpy.recarray
            Locational meta data
        attrs : dict
            Attributes to add to the meta data dataset
        """
        # pylint: disable=attribute-defined-outside-init
        self._meta = meta
        if isinstance(meta, pd.DataFrame):
            meta = to_records_array(meta)

        if ds in self.datasets:
            self.update_dset(ds, meta)
        else:
            self._create_dset(ds, meta.shape, meta.dtype, data=meta,
                              attrs=attrs)

    def _set_time_index(self, ds, time_index, attrs=None):
        """
        Write time index to disk

        Parameters
        ----------
        ds : str
            time index dataset name
        time_index : pandas.DatetimeIndex | ndarray
            Temporal index of timesteps
        attrs : dict
            Attributes to add to the meta data dataset
        """
        # pylint: disable=attribute-defined-outside-init
        self._time_index = time_index
        if isinstance(time_index, pd.DatetimeIndex):
            time_index = time_index.astype(str)
            dtype = "S{}".format(len(time_index[0]))
            time_index = np.array(time_index, dtype=dtype)

        if ds in self.datasets:
            self.update_dset(ds, time_index)
        else:
            self._create_dset(ds, time_index.shape, time_index.dtype,
                              data=time_index, attrs=attrs)

    def get_config(self, config_name):
        """
        Get SAM config

        Parameters
        ----------
        config_name : str
            Name of config

        Returns
        -------
        config : dict
            SAM config JSON as a dictionary
        """
        if 'meta' in self.datasets:
            config = json.loads(self.h5['meta'].attrs[config_name])
        else:
            config = None

        return config

    def set_configs(self, SAM_configs):
        """
        Set SAM configuration JSONs as attributes of 'meta'

        Parameters
        ----------
        SAM_configs : dict
            Dictionary of SAM configuration JSONs
        """
        if self.writable:
            for key, config in SAM_configs.items():
                if isinstance(config, dict):
                    config = json.dumps(config)

                if not isinstance(key, str):
                    key = str(key)

                self.h5['meta'].attrs[key] = config

    def _set_ds_array(self, ds_name, arr, ds_slice):
        """
        Write ds to disk

        Parameters
        ----------
        ds_name : str
            Dataset name
        arr : ndarray
            Dataset data array
        ds_slice : tuple
            Dataset slicing that corresponds to arr
        """
        if ds_name not in self.datasets:
            msg = '{} must be initialized!'.format(ds_name)
            raise HandlerRuntimeError(msg)

        dtype = self.h5[ds_name].dtype
        attrs = self.get_attrs(ds_name)
        ds_slice = parse_slice(ds_slice)
        self.h5[ds_name][ds_slice] = self._check_data_dtype(
            ds_name, arr, dtype, attrs=attrs)

    def _check_chunks(self, chunks, data=None):
        """
        Convert dataset chunk size into valid tuple based on variable array
        shape

        Parameters
        ----------
        chunks : tuple
            Desired dataset chunk size
        data : ndarray
            Dataset array being chunked

        Returns
        -------
        ds_chunks : tuple | None
            dataset chunk size
        """
        if chunks is None:
            return None

        if data is not None:
            shape = data.shape
        else:
            shape = self.shape

        if len(shape) != len(chunks):
            msg = ('Shape dimensions ({}) are not the same length as chunks '
                   '({}). Please provide a single chunk value for each '
                   'dimension!'
                    .format(shape, chunks))
            logger.error(msg)
            raise HandlerRuntimeError(msg)

        return tuple(np.min((s, s if c is None else c))
                     for s, c in zip(shape, chunks))

    def _create_dset(self, ds_name, shape, dtype, chunks=None, attrs=None,
                     data=None, replace=True):
        """
        Initialize dataset

        Parameters
        ----------
        ds_name : str
            Dataset name
        shape : tuple
            Dataset shape
        dtype : str
            Dataset numpy dtype
        chunks : tuple
            Dataset chunk size
        attrs : dict
            Dataset attributes
        data : ndarray
            Dataset data array
        replace : bool
            If previous dataset exists with the same name, it will be replaced.
        """
        ds = None
        if self.writable:
            if ds_name in self.datasets and replace:
                del self.h5[ds_name]

            elif ds_name in self.datasets:
                old_shape, old_dtype, _ = self.get_dset_properties(ds_name)
                if old_shape != shape or old_dtype != dtype:
                    e = ('Trying to create dataset "{}", but already exists '
                         'with mismatched shape and dtype. New shape/dtype '
                         'is {}/{}, previous shape/dtype is {}/{}'
                         .format(ds_name, shape, dtype, old_shape, old_dtype))
                    logger.error(e)
                    raise HandlerRuntimeError(e)

            if ds_name not in self.datasets:
                chunks = self._check_chunks(chunks, data=data)
                try:
                    ds = self.h5.create_dataset(ds_name, shape=shape,
                                                dtype=dtype, chunks=chunks)
                except Exception as e:
                    msg = ('Could not create dataset "{}" in file!'
                           .format(ds_name))
                    logger.error(msg)
                    raise IOError(msg) from e

            if attrs is not None:
                self._create_ds_attrs(ds, ds_name, attrs)

            if data is not None:
                ds[...] = data

    @staticmethod
    def _create_ds_attrs(ds, ds_name, attrs):
        """Create dataset attributes.

        Parameters
        ----------
        ds : h5py.Dataset
            Dataset object to write attributes to.
        ds_name : str
            Dataset name for logging / debugging
        attrs : dict | None
            Dataset attributes to write (None if no attributes to write).
        """
        if attrs is not None:
            for key, value in attrs.items():
                try:
                    ds.attrs[key] = value
                except Exception as e:
                    msg = ('Could not save datset "{}" attribute "{}" '
                           'to value: {}'.format(ds_name, key, value))
                    logger.error(msg)
                    raise IOError(msg) from e

    def _check_dset_shape(self, dset_name, dset_data):
        """
        Check to ensure that dataset array is of the proper shape

        Parameters
        ----------
        dset_name : str
            Dataset name being written to disk.
        dset_data : ndarray
            Dataset data array
        """
        dset_shape = dset_data.shape
        if len(dset_shape) == 1:
            possible_shapes = {}
            try:
                possible_shapes["spatial"] = (len(self.meta),)
            except ResourceKeyError:
                pass
            try:
                possible_shapes["temporal"] = (len(self.time_index),)
            except ResourceKeyError:
                pass

            if not possible_shapes:
                msg = ("Please load either 'meta' or 'time_index' before "
                       "loading a 1D dataset.")
                logger.error(msg)
                raise HandlerRuntimeError(msg)

            if dset_shape not in possible_shapes.values():
                possible_shapes_str = " or ".join(["{} {}".format(k, v)
                                                   for k, v
                                                   in possible_shapes.items()])
                msg = ('1D dataset "{}" with shape {} is not of '
                       'the proper {} shape!'
                       .format(dset_name, dset_shape, possible_shapes_str))
                logger.error(msg)
                raise HandlerValueError(msg)
        else:
            shape = self.shape
            if shape:
                if dset_shape != shape:
                    msg = ('2D dataset "{}" with shape {} is not of the '
                           'proper spatiotemporal shape: {}'
                           .format(dset_name, dset_shape, shape))
                    logger.error(msg)
                    raise HandlerValueError(msg)
            else:
                msg = ("'meta' and 'time_index' have not been loaded")
                logger.error(msg)
                raise HandlerRuntimeError(msg)

    def _add_dset(self, dset_name, data, dtype, chunks=None, attrs=None):
        """
        Write dataset to disk. Dataset it created in .h5 file and data is
        scaled if needed.

        Parameters
        ----------
        dset_name : str
            Name of dataset to be added to h5 file.
        data : ndarray
            Data to be added to h5 file.
        dtype : str
            Intended dataset datatype after scaling.
        chunks : tuple
            Chunk size for capacity factor means dataset.
        attrs : dict
            Attributes to be set. May include 'scale_factor'.
        """
        self._check_dset_shape(dset_name, data)

        data = self._check_data_dtype(dset_name, data, dtype, attrs=attrs)

        self._create_dset(dset_name, data.shape, dtype,
                          chunks=chunks, attrs=attrs, data=data)

    def update_dset(self, dset, dset_array, dset_slice=None):
        """
        Check to see if dset needs to be updated on disk
        If so write dset_array to disk

        Parameters
        ----------
        dset : str
            dataset to update
        dset_array : ndarray
            dataset array
        dset_slice : tuple
            slice of dataset to update, it None update all
        """
        if dset_slice is None:
            dset_slice = (slice(None, None, None), )

        keys = (dset, ) + dset_slice

        arr = self.__getitem__(keys)
        if not np.array_equal(arr, dset_array):
            self._set_ds_array(dset, dset_array, dset_slice)

    def write_dataset(self, dset_name, data, dtype, chunks=None, attrs=None):
        """
        Write dataset to disk. Dataset it created in .h5 file and data is
        scaled if needed.

        Parameters
        ----------
        dset_name : str
            Name of dataset to be added to h5 file.
        data : ndarray
            Data to be added to h5 file.
        dtype : str
            Intended dataset datatype after scaling.
        chunks : tuple
            Chunk size for capacity factor means dataset.
        attrs : dict
            Attributes to be set. May include 'scale_factor'.
        """
        self._add_dset(dset_name, data, dtype, chunks=chunks, attrs=attrs)

    @classmethod
    def write_profiles(cls, h5_file, meta, time_index, dset_name, profiles,
                       dtype, attrs=None, SAM_configs=None, chunks=(None, 100),
                       unscale=True, mode='w-', str_decode=True, group=None):
        """
        Write profiles to disk

        Parameters
        ----------
        h5_file : str
            Path to .h5 resource file
        meta : pandas.Dataframe
            Locational meta data
        time_index : pandas.DatetimeIndex
            Temporal timesteps
        dset_name : str
            Name of the target dataset (should identify the profiles).
        profiles : ndarray
            output result timeseries profiles
        dtype : str
            Intended dataset datatype after scaling.
        attrs : dict, optional
            Attributes to be set. May include 'scale_factor', by default None
        SAM_configs : dict, optional
            Dictionary of SAM configuration JSONs used to compute cf means,
            by default None
        chunks : tuple, optional
            Chunk size for capacity factor means dataset,
            by default (None, 100)
        unscale : bool, optional
            Boolean flag to automatically unscale variables on extraction,
            by default True
        mode : str, optional
            Mode to instantiate h5py.File instance, by default 'w-'
        str_decode : bool, optional
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read,
            by default True
        group : str, optional
            Group within .h5 resource file to open, by default None
        """
        logger.info("Saving profiles ({}) to {}".format(dset_name, h5_file))
        if profiles.shape != (len(time_index), len(meta)):
            raise HandlerValueError("Profile dimensions does not match"
                                    "'time_index' and 'meta'")
        ts = time.time()
        kwargs = {"unscale": unscale, "mode": mode, "str_decode": str_decode,
                  "group": group}
        with cls(h5_file, **kwargs) as f:
            # Save time index
            f['time_index'] = time_index
            logger.debug("\t- 'time_index' saved to disc")
            # Save meta
            f['meta'] = meta
            logger.debug("\t- 'meta' saved to disc")
            # Add SAM configurations as attributes to meta
            if SAM_configs is not None:
                f.set_configs(SAM_configs)
                logger.debug("\t- SAM configurations saved as attributes "
                             "on 'meta'")

            # Write dset to disk
            f._add_dset(dset_name, profiles, dtype,
                        chunks=chunks, attrs=attrs)
            logger.debug("\t- '{}' saved to disc".format(dset_name))

        tt = (time.time() - ts) / 60
        logger.info('{} is complete'.format(h5_file))
        logger.debug('\t- Saving to disc took {:.4f} minutes'
                     .format(tt))

    @classmethod
    def write_means(cls, h5_file, meta, dset_name, means, dtype, attrs=None,
                    SAM_configs=None, chunks=None, unscale=True, mode='w-',
                    str_decode=True, group=None):
        """
        Write means array to disk

        Parameters
        ----------
        h5_file : str
            Path to .h5 resource file
        meta : pandas.Dataframe
            Locational meta data
        dset_name : str
            Name of the target dataset (should identify the means).
        means : ndarray
            output means array.
        dtype : str
            Intended dataset datatype after scaling.
        attrs : dict, optional
            Attributes to be set. May include 'scale_factor', by default None
        SAM_configs : dict, optional
            Dictionary of SAM configuration JSONs used to compute cf means,
            by default None
        chunks : tuple, optional
            Chunk size for capacity factor means dataset, by default None
        unscale : bool, optional
            Boolean flag to automatically unscale variables on extraction,
            by default True
        mode : str, optional
            Mode to instantiate h5py.File instance, by default 'w-'
        str_decode : bool, optional
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read,
            by default True
        group : str, optional
            Group within .h5 resource file to open, by default None
        """
        logger.info("Saving means ({}) to {}".format(dset_name, h5_file))
        if len(means) != len(meta):
            msg = 'Number of means does not match meta'
            raise HandlerValueError(msg)

        ts = time.time()
        kwargs = {"unscale": unscale, "mode": mode, "str_decode": str_decode,
                  "group": group}
        with cls(h5_file, **kwargs) as f:
            # Save meta
            f['meta'] = meta
            logger.debug("\t- 'meta' saved to disc")
            # Add SAM configurations as attributes to meta
            if SAM_configs is not None:
                f.set_configs(SAM_configs)
                logger.debug("\t- SAM configurations saved as attributes "
                             "on 'meta'")

            # Write dset to disk
            f._add_dset(dset_name, means, dtype,
                        chunks=chunks, attrs=attrs)
            logger.debug("\t- '{}' saved to disc".format(dset_name))

        tt = (time.time() - ts) / 60
        logger.info('{} is complete'.format(h5_file))
        logger.debug('\t- Saving to disc took {:.4f} minutes'
                     .format(tt))

    @classmethod
    def add_dataset(cls, h5_file, dset_name, dset_data, dtype, attrs=None,
                    chunks=None, unscale=True, mode='a', str_decode=True,
                    group=None):
        """
        Add dataset to h5_file

        Parameters
        ----------
        h5_file : str
            Path to .h5 resource file
        dset_name : str
            Name of dataset to be added to h5 file
        dset_data : ndarray
            Data to be added to h5 file
        dtype : str
            Intended dataset datatype after scaling.
        attrs : dict, optional
            Attributes to be set. May include 'scale_factor', by default None
        unscale : bool, optional
            Boolean flag to automatically unscale variables on extraction,
            by default True
        mode : str, optional
            Mode to instantiate h5py.File instance, by default 'a'
        str_decode : bool, optional
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read,
            by default True
        group : str, optional
            Group within .h5 resource file to open, by default None
        """
        logger.info("Adding {} to {}".format(dset_name, h5_file))
        ts = time.time()
        kwargs = {"unscale": unscale, "mode": mode, "str_decode": str_decode,
                  "group": group}
        with cls(h5_file, **kwargs) as f:
            f._add_dset(dset_name, dset_data, dtype,
                        chunks=chunks, attrs=attrs)

        tt = (time.time() - ts) / 60
        logger.info('{} added'.format(dset_name))
        logger.debug('\t- Saving to disc took {:.4f} minutes'
                     .format(tt))

    @classmethod
    def init_h5(cls, h5_file, dsets, shapes, attrs, chunks, dtypes,
                meta, time_index=None, configs=None, unscale=True, mode='w',
                str_decode=True, group=None, run_attrs=None):
        """Init a full output file with the final intended shape without data.

        Parameters
        ----------
        h5_file : str
            Full h5 output filepath.
        dsets : list
            List of strings of dataset names to initialize (does not include
            meta or time_index).
        shapes : dict
            Dictionary of dataset shapes (keys correspond to dsets).
        attrs : dict
            Dictionary of dataset attributes (keys correspond to dsets).
        chunks : dict
            Dictionary of chunk tuples (keys correspond to dsets).
        dtypes : dict
            dictionary of numpy datatypes (keys correspond to dsets).
        meta : pd.DataFrame
            Full meta data.
        time_index : pd.datetimeindex | None
            Full pandas datetime index. None implies that only 1D results
            (no site profiles) are being written.
        configs : dict | None
            Optional input configs to set as attr on meta.
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        mode : str
            Mode to instantiate h5py.File instance
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        group : str
            Group within .h5 resource file to open
        run_attrs : dict | NoneType
            Runtime attributes (args, kwargs) to add as global (file)
            attributes
        """

        logger.debug("Initializing output file: {}".format(h5_file))
        kwargs = {"unscale": unscale, "mode": mode, "str_decode": str_decode,
                  "group": group}
        with cls(h5_file, **kwargs) as f:
            if run_attrs is not None:
                f.run_attrs = run_attrs

            f['meta'] = meta

            if time_index is not None:
                f['time_index'] = time_index

            for dset in dsets:
                if dset not in ('meta', 'time_index'):
                    # initialize each dset to disk
                    f._create_dset(dset, shapes[dset], dtypes[dset],
                                   chunks=chunks[dset], attrs=attrs[dset])

            if configs is not None:
                f.set_configs(configs)
                logger.debug("\t- Configurations saved as attributes "
                             "on 'meta'")

        logger.debug('Output file has been initialized.')



# -*- coding: utf-8 -*-
"""
Temporal Statistics Extraction
"""
from concurrent.futures import as_completed
import gc
import logging
import numpy as np
import os
import pandas as pd

from rex.resource import Resource
from rex.utilities.bc_utils import (sample_q_linear, sample_q_log,
                                    sample_q_invlog)
from rex.utilities.execution import SpawnProcessPool
from rex.utilities.loggers import log_mem, log_versions, create_dirs
from rex.utilities.utilities import get_lat_lon_cols, slice_sites

logger = logging.getLogger(__name__)


def circular_mean(data, weights=None, degrees=True, axis=0,
                  exponential_weights=True):
    """
    Computed the ciruclar average. if provided compute the weighed average
    with the given weights. For example, if averaging wind direction with wind
    speed as weights, wind directions that occur at higher wind speeds will
    have a larger weight of the final mean value.

    Parameters
    ----------
    data : ndarray
        Data to average
    weights : ndarray, optional
        Weights to apply to data during averaging, must be of the same
        shape as data, by default None
    degree : bool, optional
        Flag indicating that data is in degrees and needs to be converted
        to/from radians during averaging. By default True
    axis : int, optional
        Axis to compute average along, by default 0 which will produce
        site averages
    norm_weights: : bool, optional
        Flag to normalize weights, by default True
    exponential_weights : bool
        Flag to convert weights to exponential, by default True

    Returns
    -------
    mean : ndarray
        Weighted circular mean along the given axis
    """
    if degrees:
        data = np.radians(data, dtype=np.float32)

    sin = np.sin(data)
    cos = np.cos(data)
    if weights is None:
        sin = np.nanmean(sin, axis=axis)
        cos = np.nanmean(cos, axis=axis)
    else:
        if exponential_weights:
            weights = np.exp(weights)

        if weights.shape != data.shape:
            msg = ('The shape of weights {} does not match the shape of the '
                   'data {} to which it is to be applied!'
                   .format(weights.shape, data.shape))
            logger.error(msg)
            raise RuntimeError(msg)

        n_weights = np.expand_dims(np.nansum(weights, axis=axis), axis)
        sin = np.nansum(sin * weights, axis=axis) / n_weights
        cos = np.nansum(cos * weights, axis=axis) / n_weights

    mean = np.arctan2(sin, cos)
    if degrees:
        mean = np.degrees(mean)
        mask = mean < 0
        if isinstance(mask, np.ndarray):
            mean[mask] += 360
        elif mask:
            mean += 360

    return mean


def cdf(data, n_samples=50, sampling='linear', log_base=10, decimals=None):
    """Get a number of x-values that define a CDF for the input data.

    Parameters
    ----------
    data : np.ndarray
        1D array of data to make a CDF for
    n_samples : int
        Number of points to fit the CDF
    sampling : str
        Option for quantile sampling (see sampling functions in
        ``rex.utilities.bc_utils``), e.g., how to sample the y-axis of the
        distribution. "linear" will do even spacing, "log" will concentrate
        samples near quantile=0, and "invlog" will concentrate samples near
        quantile=1
    log_base : int | float
        Log base value if sampling is "log" or "invlog". A higher value will
        concentrate more samples at the extreme sides of the distribution.
    decimals : int | None
        Precision to round output to (see docstring for np.round). None will
        not round outputs (default).

    Returns
    -------
    x_values : np.ndarray
        1D array of values with shape (n_samples,). Each value is in the same
        units as the input data argument. The x_values[0] is the minimum value
        of data (0th percentile) and x_values[-1] is the maximum
        (100th percentile). The values are spaced in quantile space (y-axis of
        the CDF) according to the sampling option (e.g., evenly spaced if
        sampling='linear').
    """

    nan_mask = np.isnan(data)
    if nan_mask.all():
        return np.zeros(n_samples)

    sampling = sampling.casefold()
    if sampling == 'linear':
        quantiles = sample_q_linear(n_samples)
    elif sampling == 'log':
        quantiles = sample_q_log(n_samples, log_base)
    elif sampling == 'invlog':
        quantiles = sample_q_invlog(n_samples, log_base)
    else:
        msg = ('sampling option must be linear, log, or invlog, but received: '
               '{}'.format(sampling))
        logger.error(msg)
        raise KeyError(msg)

    x_values = np.interp(quantiles, np.linspace(0, 1, len(data[~nan_mask])),
                         sorted(data[~nan_mask]))

    msg = (f'First and last x-value points defining the CDF '
           '({x_values[0]}, {x_values[-1]}) '
           f'were not the min and max data values '
           f'({np.nanmin(data)}, {np.nanmin(data)}).')
    assert x_values[0] == np.nanmin(data), msg
    assert x_values[-1] == np.nanmax(data), msg

    if decimals is not None:
        x_values = np.round(x_values, decimals=decimals)

    return x_values


class TemporalStats:
    """
    Temporal Statistics from Resource Data
    """
    STATS = {'mean': {'func': np.nanmean, 'kwargs': {'axis': 0}},
             'median': {'func': np.nanmedian, 'kwargs': {'axis': 0}},
             'std': {'func': np.nanstd, 'kwargs': {'axis': 0}}}

    def __init__(self, res_h5, statistics='mean', res_cls=Resource,
                 hsds=False):
        """
        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        statistics : str | tuple | dict, optional
            Statistics to extract, either a key or tuple of keys in
            cls.STATS, or a dictionary of the form
            {'stat_name': {'func': *, 'kwargs: {**}}},
            by default 'mean'
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        """
        log_versions(logger)
        self._res_h5 = res_h5
        self._stats = None
        self.statistics = statistics

        self._res_cls = res_cls
        self._hsds = hsds

        with res_cls(res_h5, hsds=self._hsds) as f:
            self._time_index = f.time_index
            self._meta = f.meta

    @property
    def res_h5(self):
        """
        Path to resource h5 file(s)

        Returns
        -------
        str
        """
        return self._res_h5

    @property
    def statistics(self):
        """
        Dictionary of statistic functions/kwargs to run

        Returns
        -------
        dict
        """
        return self._stats

    @statistics.setter
    def statistics(self, statistics):
        """
         Statistics to extract, either a key or tuple of keys in
        cls.STATS, or a dictionary of the form
        {'stat_name': {'func': *, 'kwargs: {**}}}

        Parameters
        ----------
        statistics : dict
        """
        self._stats = self._check_stats(statistics)

    @property
    def res_cls(self):
        """
        Resource class to use to access res_h5

        Returns
        -------
        Class
        """
        return self._res_cls

    @property
    def time_index(self):
        """
        Resource Datetimes

        Returns
        -------
        pandas.DatetimeIndex
        """
        return self._time_index

    @property
    def meta(self):
        """
        Resource meta-data table

        Returns
        -------
        pandas.DataFrame
        """
        return self._meta

    @property
    def lat_lon(self):
        """
        Resource (lat, lon) coordinates

        Returns
        -------
        pandas.DataFrame
        """
        lat_lon_cols = get_lat_lon_cols(self.meta)

        return self.meta[lat_lon_cols]

    @staticmethod
    def _format_grp_names(grp_names):
        """
        Format groupby index values

        Parameters
        ----------
        grp_names : list
            Group by index values, these correspond to each unique group in
            the groupby

        Returns
        -------
        out : ndarray
            2D array of grp index values properly formatted as strings
        """
        month_map = {1: 'Jan', 2: 'Feb', 3: 'Mar', 4: 'Apr', 5: 'May',
                     6: 'Jun', 7: 'Jul', 8: 'Aug', 9: 'Sep', 10: 'Oct',
                     11: 'Nov', 12: 'Dec'}

        # pylint: disable=unnecessary-lambda-assignment, unnecessary-lambda
        year = lambda s: "{}".format(s)
        month = lambda s: "{}".format(month_map[s])
        hour = lambda s: "{:02d}:00UTC".format(s)

        grp_names = np.array(grp_names).T
        if len(grp_names.shape) == 1:
            grp_names = np.expand_dims(grp_names, 0)

        out = []
        for grp_i in grp_names:  # pylint: disable=not-an-iterable
            grp_max = grp_i.max()
            if grp_max <= 12:
                out.append(list(map(month, grp_i)))
            elif grp_max <= 23:
                out.append(list(map(hour, grp_i)))
            else:
                out.append(list(map(year, grp_i)))

        return np.array(out).T

    @classmethod
    def _create_names(cls, groups, stats):
        """
        Generate statistics names

        Parameters
        ----------
        groups : list
            List of group names, some combination of year, month, hour
        stats : list
            Statistics to be computed

        Returns
        -------
        columns_map : dict
            Dictionary of column names to use for each statistic
        """
        group_names = cls._format_grp_names(groups)

        columns_map = {}
        for s in stats:
            # pylint: disable=not-an-iterable
            cols = ['{}_{}'.format('-'.join(n), s) for n
                    in group_names]
            columns_map[s] = cols

        return columns_map

    @staticmethod
    def _compute_weighted_stats(func, res_data, weights, column_names,
                                **kwargs):
        """
        Computed the weighted means using given function and kwargs

        Parameters
        ----------
        func : object
            Function to use to compute the weighted means
        res_data : pandas.DataFrame | pandas.GroupBy
            Resource data to compute weighted stats from
        weights : pandas.DataFrame | pandas.GroupBy
            Weights to use for weighted stats calculation
            Column names based on group by names, by default None
        column_names : list | str
            Either the state name or the list of output stat names,
            used out output column names.
        kwargs : dict
            Function kwargs
        """
        if isinstance(column_names, list):
            s_data = {}
            for c_name, (grp_name, res_grp) in zip(column_names, res_data):
                if weights is not None:
                    grp_w = weights.get_group(grp_name[0])
                else:
                    grp_w = None

                grp_s = func(res_grp, weights=grp_w, **kwargs)
                if grp_s.shape[0] == 1:
                    grp_s = grp_s[0]

                s_data[c_name] = grp_s

            s_data = pd.DataFrame(s_data)
        else:
            s_data = func(res_data, weights=weights, **kwargs)
            if s_data.shape[0] == 1:
                s_data = s_data[0]

            s_data = pd.DataFrame(s_data.flatten(), columns=[column_names])

        return s_data

    @classmethod
    def _compute_stats(cls, res_data, statistics, diurnal=False, month=False):
        """
        Compute desired stats for desired time intervals from res_data

        Parameters
        ----------
        res_data : pandas.DataFrame
            DataFrame or resource data. Index is time_index, columns are sites
        statistics : dict
            Dictionary of statistic functions/kwargs to run
        diurnal : bool, optional
            Extract diurnal stats, by default False
        month : bool, optional
            Extract monthly stats, by default False

        Returns
        -------
        res_stats : pandas.DataFrame
            DataFrame of desired statistics at desired time intervals
        """
        groupby = []
        column_names = None
        if month:
            groupby.append(res_data.index.month)

        if diurnal:
            groupby.append(res_data.index.hour)

        if groupby:
            res_data = res_data.groupby(groupby)
            column_names = cls._create_names(list(res_data.groups),
                                             list(statistics))

        res_stats = []
        for name, stat in statistics.copy().items():
            func = stat['func']
            kwargs = stat.get('kwargs', {}).copy()
            if name.lower().startswith('weight'):
                weights = kwargs.pop('weights').copy()
                if groupby:
                    weights = weights.groupby(groupby)
                    weight_names = column_names[name]
                else:
                    weight_names = name

                s_data = cls._compute_weighted_stats(func, res_data, weights,
                                                     weight_names, **kwargs)
            else:
                axis = kwargs.pop('axis', 0)
                s_data = res_data.aggregate(func, axis=axis, **kwargs)

                if groupby:
                    columns = column_names[name]
                    s_data = s_data.T
                    s_data.columns = columns
                elif not isinstance(s_data, pd.DataFrame):
                    s_data = s_data.to_frame(name=name)
                elif isinstance(s_data, pd.DataFrame) and len(s_data) > 1:
                    # e.g., if func is scipy.stats.beta.fit(), this collapses
                    # multiple output parameters into list
                    s_data['name'] = name
                    s_data = s_data.groupby('name').agg(list).T

            res_stats.append(s_data)

        res_stats = pd.concat(res_stats, axis=1)

        return res_stats

    @staticmethod
    def _create_index(sites_slice):
        """
        Create index from site slice

        Parameters
        ----------
        sites_slice : slice | list | ndarray
            Sites to build index from

        Returns
        -------
        idx : list
            site gids
        """
        if isinstance(sites_slice, slice) and sites_slice.stop:
            idx = list(range(*sites_slice.indices(sites_slice.stop)))
        elif isinstance(sites_slice, (list, np.ndarray)):
            idx = sites_slice

        return idx

    @staticmethod
    def _extract_weights(res, weights_dsets, sites_slice, time_index):
        """
        Extract weights datasets from resource and combine into weights
        to use for weighted stats

        Parameters
        ----------
        res : rex.Resource
            Open Resource class or sub-class to extract datasets from
        weights_dsets : str | list | tuple
            List of weight(s) datasets to extract and combine
        sites_slice : slice
            Subslice of sites to extract weights for
        time_index : pandas.DatatimeIndex
            Resource DatetimeIndex, needed to output DataFrame Index

        Returns
        -------
        weights : pandas.DataFrame
            Weights DataFrame to match res_data
        """
        if not isinstance(weights_dsets, (list, tuple)):
            weights_dsets = [weights_dsets]

        weights = None
        for dset in weights_dsets:
            if weights is None:
                weights = res[dset, :, sites_slice]
            else:
                weights *= res[dset, :, sites_slice]

        return pd.DataFrame(weights, index=time_index)

    @classmethod
    def _extract_stats(cls, res_h5, statistics, dataset, res_cls=Resource,
                       hsds=False, time_index=None, sites_slice=None,
                       diurnal=False, month=False, combinations=False,
                       mask_zeros=False):
        """
        Extract stats for given dataset, sites, and temporal extent

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        statistics : dict
            Statistics to extract a dictionary of the form
            {'stat_name': {'func': *, 'kwargs: {**}}}
        dataset : str
            Dataset to extract stats for
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        time_index : pandas.DatatimeIndex | None, optional
            Resource DatetimeIndex, if None extract from res_h5,
            by default None
        sites_slice : slice | None, optional
            Sites to extract, if None all, by default None
        diurnal : bool, optional
            Extract diurnal stats, by default False
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False
        mask_zeros : bool
            Flag to only calculate stats when all data is > 0 (useful for
            global horizontal irradiance).

        Returns
        -------
        res_stats : pandas.DataFrame
            DataFrame of desired statistics at desired time intervals
        """
        if sites_slice is None:
            sites_slice = slice(None, None, None)

        with res_cls(res_h5, hsds=hsds) as f:
            if time_index is None:
                time_index = f.time_index

            res_data = pd.DataFrame(f[dataset, :, sites_slice],
                                    index=time_index)

            if mask_zeros:
                res_data[(res_data == 0)] = np.nan

            for s, s_dict in statistics.items():
                weights = s_dict.get('kwargs', {}).get('weights')
                if weights is not None:
                    weights = cls._extract_weights(f, weights, sites_slice,
                                                   time_index)
                    statistics[s]['kwargs']['weights'] = weights

        if combinations:
            res_stats = [cls._compute_stats(res_data, statistics)]
            if month:
                res_stats.append(cls._compute_stats(res_data, statistics,
                                                    month=True))

            if diurnal:
                res_stats.append(cls._compute_stats(res_data, statistics,
                                                    diurnal=True))
            if month and diurnal:
                res_stats.append(cls._compute_stats(res_data, statistics,
                                                    month=True, diurnal=True))

            res_stats = pd.concat(res_stats, axis=1)
        else:
            res_stats = cls._compute_stats(res_data, statistics,
                                           diurnal=diurnal, month=month)

        res_stats.index = cls._create_index(sites_slice)
        res_stats.index.name = 'gid'

        return res_stats

    def _get_slices(self, dataset, sites=None, chunks_per_slice=5):
        """
        Get slices to extract

        Parameters
        ----------
        dataset : str
            Dataset to extract data from
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
            (sites is synonymous with gids aka spatial indices)
        chunks_per_slice : int, optional
            Number of chunks to extract in each slice, by default 5

        Returns
        -------
        slices : list
            List of slices to extract
        """
        with self.res_cls(self.res_h5) as f:
            shape, _, chunks = f.get_dset_properties(dataset)

        if len(shape) != 2:
            msg = ('Cannot extract temporal stats for dataset {}, as it is '
                   'not a timeseries dataset!'.format(dataset))
            logger.error(msg)
            raise RuntimeError(msg)

        slices = slice_sites(shape, chunks, sites=sites,
                             chunks_per_slice=chunks_per_slice)

        return slices

    def _check_stats(self, statistics):
        """
        check desired statistics to make sure inputs are valid

        Parameters
        ----------
        statistics : str | tuple | dict
            Statistics to extract, either a key or tuple of keys in
            cls.STATS, or a dictionary of the form
            {'stat_name': {'func': *, 'kwargs: {**}}}

        Returns
        -------
        stats : dict
            Dictionary of statistic functions/kwargs to run
        """
        if isinstance(statistics, str):
            statistics = (statistics, )

        if isinstance(statistics, (tuple, list)):
            statistics = {s: self.STATS[s] for s in statistics}

        for stat in statistics.values():
            msg = 'A "func"(tion) must be provided for each statistic'
            assert 'func' in stat, msg
            if 'kwargs' in stat:
                msg = 'statistic function kwargs must be a dictionary '
                assert isinstance(stat['kwargs'], dict), msg

        return statistics

    def compute_statistics(self, dataset, sites=None, diurnal=False,
                           month=False, combinations=False, max_workers=None,
                           chunks_per_worker=5, lat_lon_only=True,
                           mask_zeros=False):
        """
        Compute statistics

        Parameters
        ----------
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
            (sites is synonymous with gids aka spatial indices)
        diurnal : bool, optional
            Extract diurnal stats, by default False
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        mask_zeros : bool
            Flag to only calculate stats when all data is > 0 (useful for
            global horizontal irradiance).

        Returns
        -------
        res_stats : pandas.DataFrame
            DataFrame of desired statistics at desired time intervals
        """
        if max_workers is None:
            max_workers = os.cpu_count()

        slices = self._get_slices(dataset, sites,
                                  chunks_per_slice=chunks_per_worker)

        if len(slices) == 1:
            max_workers = 1

        if max_workers > 1:
            msg = ('Extracting {} for {} in parallel using {} workers'
                   .format(list(self.statistics), dataset, max_workers))
            logger.info(msg)

            loggers = [__name__, 'rex']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                for sites_slice in slices:
                    future = exe.submit(self._extract_stats,
                                        self.res_h5, self.statistics, dataset,
                                        res_cls=self.res_cls,
                                        hsds=self._hsds,
                                        time_index=self.time_index,
                                        sites_slice=sites_slice,
                                        diurnal=diurnal,
                                        month=month,
                                        combinations=combinations,
                                        mask_zeros=mask_zeros)
                    futures.append(future)

                res_stats = []
                for i, future in enumerate(as_completed(futures)):
                    res_stats.append(future.result())
                    logger.debug('Completed {} out of {} workers'
                                 .format((i + 1), len(futures)))
        else:
            msg = ('Extracting {} for {} in serial'
                   .format(self.statistics.keys(), dataset))
            logger.info(msg)
            res_stats = []
            for i, sites_slice in enumerate(slices):
                res_stats.append(self._extract_stats(
                    self.res_h5, self.statistics, dataset,
                    res_cls=self.res_cls, hsds=self._hsds,
                    time_index=self.time_index, sites_slice=sites_slice,
                    diurnal=diurnal, month=month,
                    combinations=combinations,
                    mask_zeros=mask_zeros))
                logger.debug('Completed {} out of {} sets of sites'
                             .format((i + 1), len(slices)))

        gc.collect()
        log_mem(logger)
        res_stats = pd.concat(res_stats)

        if lat_lon_only:
            meta = self.lat_lon
        else:
            meta = self.meta

        res_stats = meta.join(res_stats.sort_index(), how='inner')

        return res_stats

    def full_stats(self, dataset, sites=None, max_workers=None,
                   chunks_per_worker=5, lat_lon_only=True, mask_zeros=False):
        """
        Compute stats for entire temporal extent of file

        Parameters
        ----------
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
            (sites is synonymous with gids aka spatial indices)
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        mask_zeros : bool
            Flag to only calculate stats when all data is > 0 (useful for
            global horizontal irradiance).

        Returns
        -------
        full_stats : pandas.DataFrame
            DataFrame of statistics for the entire temporal extent of file
        """
        full_stats = self.compute_statistics(
            dataset, sites=sites,
            max_workers=max_workers,
            chunks_per_worker=chunks_per_worker,
            lat_lon_only=lat_lon_only,
            mask_zeros=mask_zeros)

        return full_stats

    def monthly_stats(self, dataset, sites=None, max_workers=None,
                      chunks_per_worker=5, lat_lon_only=True,
                      mask_zeros=False):
        """
        Compute monthly stats

        Parameters
        ----------
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
            (sites is synonymous with gids aka spatial indices)
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        mask_zeros : bool
            Flag to only calculate stats when all data is > 0 (useful for
            global horizontal irradiance).

        Returns
        -------
        monthly_stats : pandas.DataFrame
            DataFrame of monthly statistics
        """
        monthly_stats = self.compute_statistics(
            dataset, sites=sites, month=True,
            max_workers=max_workers,
            chunks_per_worker=chunks_per_worker,
            lat_lon_only=lat_lon_only,
            mask_zeros=mask_zeros)

        return monthly_stats

    def diurnal_stats(self, dataset, sites=None, max_workers=None,
                      chunks_per_worker=5, lat_lon_only=True,
                      mask_zeros=False):
        """
        Compute diurnal stats

        Parameters
        ----------
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
            (sites is synonymous with gids aka spatial indices)
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        mask_zeros : bool
            Flag to only calculate stats when all data is > 0 (useful for
            global horizontal irradiance).

        Returns
        -------
        diurnal_stats : pandas.DataFrame
            DataFrame of diurnal statistics
        """
        diurnal_stats = self.compute_statistics(
            dataset, sites=sites, diurnal=True,
            max_workers=max_workers,
            chunks_per_worker=chunks_per_worker,
            lat_lon_only=lat_lon_only,
            mask_zeros=mask_zeros)

        return diurnal_stats

    def monthly_diurnal_stats(self, dataset, sites=None,
                              max_workers=None, chunks_per_worker=5,
                              lat_lon_only=True, mask_zeros=False):
        """
        Compute monthly-diurnal stats

        Parameters
        ----------
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
            (sites is synonymous with gids aka spatial indices)
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        mask_zeros : bool
            Flag to only calculate stats when all data is > 0 (useful for
            global horizontal irradiance).

        Returns
        -------
        monthly_diurnal_stats : pandas.DataFrame
            DataFrame of monthly-diurnal statistics
        """
        diurnal_stats = self.compute_statistics(
            dataset, sites=sites, month=True, diurnal=True,
            max_workers=max_workers,
            chunks_per_worker=chunks_per_worker,
            lat_lon_only=lat_lon_only,
            mask_zeros=mask_zeros)

        return diurnal_stats

    def all_stats(self, dataset, sites=None, max_workers=None,
                  chunks_per_worker=5, lat_lon_only=True, mask_zeros=False):
        """
        Compute annual, monthly, monthly-diurnal, and diurnal stats

        Parameters
        ----------
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
            (sites is synonymous with gids aka spatial indices)
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        mask_zeros : bool
            Flag to only calculate stats when all data is > 0 (useful for
            global horizontal irradiance).

        Returns
        -------
        all_diurnal_stats : pandas.DataFrame
            DataFrame of temporal statistics
        """
        all_stats = self.compute_statistics(
            dataset, sites=sites, month=True, diurnal=True, combinations=True,
            max_workers=max_workers,
            chunks_per_worker=chunks_per_worker,
            lat_lon_only=lat_lon_only,
            mask_zeros=mask_zeros)

        return all_stats

    def save_stats(self, res_stats, out_path):
        """
        Save statistics to disk

        Parameters
        ----------
        res_stats : pandas.DataFrame
            Table of statistics to save
        out_path : str
            Directory, .csv, or .json path to save statistics too
        """
        if os.path.isdir(out_path):
            out_fpath = os.path.splitext(os.path.basename(self.res_h5))[0]
            out_fpath = os.path.join(out_path, out_fpath + '.csv')
        else:
            out_fpath = out_path

        # Drop any wild card values
        out_fpath = out_fpath.replace('*', '')

        out_dir = os.path.dirname(out_fpath)
        create_dirs(out_dir)

        logger.info('Writing temporal statistics to {}'.format(out_fpath))
        if out_fpath.endswith('.csv'):
            res_stats.to_csv(out_fpath)
        elif out_fpath.endswith('.json'):
            res_stats.to_json(out_fpath)
        else:
            msg = ("Cannot save statistics, expecting a directory, .csv, or "
                   ".json path, but got: {}".format(out_path))
            logger.error(msg)
            raise OSError(msg)

    @classmethod
    def run(cls, res_h5, dataset, sites=None, statistics='mean',
            diurnal=False, month=False, combinations=False,
            res_cls=Resource, hsds=False, max_workers=None,
            chunks_per_worker=5, lat_lon_only=True, mask_zeros=False,
            out_path=None):
        """
        Compute temporal stats, by default full temporal extent stats

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
            (sites is synonymous with gids aka spatial indices)
        statistics : str | tuple | dict, optional
            Statistics to extract, either a key or tuple of keys in
            cls.STATS, or a dictionary of the form
            {'stat_name': {'func': *, 'kwargs: {**}}},
            by default 'mean'
        diurnal : bool, optional
            Extract diurnal stats, by default False
        month : bool, optional
            Extract monthly stats, by default False
        combinations : bool, optional
            Extract all combinations of temporal stats, by default False
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        mask_zeros : bool
            Flag to only calculate stats when all data is > 0 (useful for
            global horizontal irradiance).
        out_path : str, optional
            Directory, .csv, or .json path to save statistics too,
            by default None

        Returns
        -------
        out_stats : pandas.DataFrame
            DataFrame of resource statistics
        """
        logger.info('Computing temporal stats for {} in {}'
                    .format(dataset, res_h5))
        logger.debug('Computing {} using:'
                     '\n-diurnal={}'
                     '\n-month={}'
                     '\n-combinations={}'
                     '\n-max workers={}'
                     '\n-chunks per worker={}'
                     '\n-output lat lons only={}'
                     .format(statistics, diurnal, month, combinations,
                             max_workers, chunks_per_worker, lat_lon_only))
        res_stats = cls(res_h5, statistics=statistics, res_cls=res_cls,
                        hsds=hsds)
        out_stats = res_stats.compute_statistics(
            dataset, sites=sites,
            diurnal=diurnal, month=month, combinations=combinations,
            max_workers=max_workers, chunks_per_worker=chunks_per_worker,
            lat_lon_only=lat_lon_only, mask_zeros=mask_zeros)
        if out_path is not None:
            res_stats.save_stats(out_stats, out_path)

        return out_stats

    @classmethod
    def monthly(cls, res_h5, dataset, sites=None, statistics='mean',
                res_cls=Resource, hsds=False, max_workers=None,
                chunks_per_worker=5, lat_lon_only=True, mask_zeros=False,
                out_path=None):
        """
        Compute monthly stats

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
            (sites is synonymous with gids aka spatial indices)
        statistics : str | tuple | dict, optional
            Statistics to extract, either a key or tuple of keys in
            cls.STATS, or a dictionary of the form
            {'stat_name': {'func': *, 'kwargs: {**}}},
            by default 'mean'
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        mask_zeros : bool
            Flag to only calculate stats when all data is > 0 (useful for
            global horizontal irradiance).
        out_path : str, optional
            Directory, .csv, or .json path to save statistics too,
            by default None

        Returns
        -------
        monthly_stats : pandas.DataFrame
            DataFrame of monthly statistics
        """
        monthly_stats = cls.run(res_h5, dataset, sites=sites,
                                statistics=statistics, diurnal=False,
                                month=True, combinations=False,
                                res_cls=res_cls, hsds=hsds,
                                max_workers=max_workers,
                                chunks_per_worker=chunks_per_worker,
                                lat_lon_only=lat_lon_only,
                                mask_zeros=mask_zeros,
                                out_path=out_path)

        return monthly_stats

    @classmethod
    def diurnal(cls, res_h5, dataset, sites=None, statistics='mean',
                res_cls=Resource, hsds=False, max_workers=None,
                chunks_per_worker=5, lat_lon_only=True, mask_zeros=False,
                out_path=None):
        """
        Compute diurnal stats

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
            (sites is synonymous with gids aka spatial indices)
        statistics : str | tuple | dict, optional
            Statistics to extract, either a key or tuple of keys in
            cls.STATS, or a dictionary of the form
            {'stat_name': {'func': *, 'kwargs: {**}}},
            by default 'mean'
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        mask_zeros : bool
            Flag to only calculate stats when all data is > 0 (useful for
            global horizontal irradiance).
        out_path : str, optional
            Directory, .csv, or .json path to save statistics too,
            by default None

        Returns
        -------
        diurnal_stats : pandas.DataFrame
            DataFrame of diurnal statistics
        """
        diurnal_stats = cls.run(res_h5, dataset, sites=sites,
                                statistics=statistics, diurnal=True,
                                month=False, combinations=False,
                                res_cls=res_cls, hsds=hsds,
                                max_workers=max_workers,
                                chunks_per_worker=chunks_per_worker,
                                lat_lon_only=lat_lon_only,
                                mask_zeros=mask_zeros,
                                out_path=out_path)

        return diurnal_stats

    @classmethod
    def monthly_diurnal(cls, res_h5, dataset, sites=None,
                        statistics='mean', res_cls=Resource, hsds=False,
                        max_workers=None, chunks_per_worker=5,
                        lat_lon_only=True, mask_zeros=False, out_path=None):
        """
        Compute monthly-diurnal stats

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
            (sites is synonymous with gids aka spatial indices)
        statistics : str | tuple | dict, optional
            Statistics to extract, either a key or tuple of keys in
            cls.STATS, or a dictionary of the form
            {'stat_name': {'func': *, 'kwargs: {**}}},
            by default 'mean'
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        mask_zeros : bool
            Flag to only calculate stats when all data is > 0 (useful for
            global horizontal irradiance).
        out_path : str, optional
            Directory, .csv, or .json path to save statistics too,
            by default None

        Returns
        -------
        monthly_diurnal_stats : pandas.DataFrame
            DataFrame of monthly-diurnal statistics
        """
        monthly_diurnal_stats = cls.run(res_h5, dataset, sites=sites,
                                        statistics=statistics, diurnal=True,
                                        month=True, combinations=False,
                                        res_cls=res_cls, hsds=hsds,
                                        max_workers=max_workers,
                                        chunks_per_worker=chunks_per_worker,
                                        lat_lon_only=lat_lon_only,
                                        mask_zeros=mask_zeros,
                                        out_path=out_path)

        return monthly_diurnal_stats

    @classmethod
    def all(cls, res_h5, dataset, sites=None, statistics='mean',
            res_cls=Resource, hsds=False, max_workers=None,
            chunks_per_worker=5, lat_lon_only=True, mask_zeros=False,
            out_path=None):
        """
        Compute annual, monthly, monthly-diurnal, and diurnal stats

        Parameters
        ----------
        res_h5 : str
            Path to resource h5 file(s)
        dataset : str
            Dataset to extract stats for
        sites : list | slice, optional
            Subset of sites to extract, by default None or all sites
            (sites is synonymous with gids aka spatial indices)
        statistics : str | tuple | dict, optional
            Statistics to extract, either a key or tuple of keys in
            cls.STATS, or a dictionary of the form
            {'stat_name': {'func': *, 'kwargs: {**}}},
            by default 'mean'
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        res_cls : Class, optional
            Resource class to use to access res_h5, by default Resource
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        max_workers : None | int, optional
            Number of workers to use, if 1 run in serial, if None use all
            available cores, by default None
        chunks_per_worker : int, optional
            Number of chunks to extract on each worker, by default 5
        lat_lon_only : bool, optional
            Only append lat, lon coordinates to stats, by default True
        mask_zeros : bool
            Flag to only calculate stats when all data is > 0 (useful for
            global horizontal irradiance).
        out_path : str, optional
            Directory, .csv, or .json path to save statistics too,
            by default None

        Returns
        -------
        all_stats : pandas.DataFrame
            DataFrame of temporal statistics
        """
        all_stats = cls.run(res_h5, dataset, sites=sites,
                            statistics=statistics, diurnal=True,
                            month=True, combinations=True,
                            res_cls=res_cls, hsds=hsds,
                            max_workers=max_workers,
                            chunks_per_worker=chunks_per_worker,
                            lat_lon_only=lat_lon_only, out_path=out_path,
                            mask_zeros=mask_zeros)

        return all_stats



# -*- coding: utf-8 -*-
"""
Classes to handle resource data
"""
import os
from abc import ABC, abstractmethod
from warnings import warn

import dateutil
import h5py
import numpy as np
import pandas as pd

from rex.sam_resource import SAMResource
from rex.utilities.exceptions import ResourceKeyError, ResourceRuntimeError
from rex.utilities.parse_keys import parse_keys, parse_slice
from rex.utilities.utilities import check_tz, get_lat_lon_cols


class BaseDatasetIterable(ABC):
    """Base class for file that is iterable over datasets. """

    @property
    @abstractmethod
    def datasets(self):
        """iterable: Datasets available in file. """

    def __iter__(self):
        return iter(self.datasets)


class ResourceDataset:
    """
    h5py.Dataset wrapper for Resource .h5 files
    """

    def __init__(self, ds, scale_attr='scale_factor', add_attr='add_offset',
                 unscale=True):
        """
        Parameters
        ----------
        ds : h5py.dataset
            Open .h5 dataset instance to extract data from
        scale_attr : str | list | optional
            Name of scale factor attribute, by default 'scale_factor'. Can also
            be a prioritized list of scale factor names.
        add_attr : str | list | optional
            Name of add offset attribute, by default 'add_offset'. Can also
            be a prioritized list of add offset names.
        unscale : bool, optional
            Flag to unscale dataset data, by default True
        """
        self._ds = ds

        self._scale_factor = self._parse_scale_add_attrs(scale_attr, 1)
        self._adder = self._parse_scale_add_attrs(add_attr, 0)

        self._unscale = unscale
        if self._scale_factor == 1 and self._adder == 0:
            self._unscale = False

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self.ds.name)

        return msg

    def __getitem__(self, ds_slice):
        ds_slice = parse_slice(ds_slice)

        return self._get_ds_slice(ds_slice)

    @property
    def ds(self):
        """
        Open Dataset instance

        Returns
        -------
        h5py(d).Dataset
        """
        return self._ds

    @property
    def shape(self):
        """
        Dataset shape

        Returns
        -------
        tuple
        """
        return self.ds.shape

    @property
    def size(self):
        """
        Dataset size

        Returns
        -------
        int
        """
        return self.ds.size

    @property
    def dtype(self):
        """
        Dataset dtype

        Returns
        -------
        str | numpy.dtype
        """
        return self.ds.dtype

    @property
    def chunks(self):
        """
        Dataset chunk size

        Returns
        -------
        tuple
        """
        chunks = self.ds.chunks
        if isinstance(chunks, dict):
            chunks = tuple(chunks.get('dims', None))

        return chunks

    @property
    def scale_factor(self):
        """
        Dataset scale factor

        Returns
        -------
        float
        """
        return self._scale_factor

    @property
    def adder(self):
        """
        Dataset add offset

        Returns
        -------
        float
        """
        return self._adder

    def _parse_scale_add_attrs(self, attr, default):
        """Get the scale and add offset factors using one or more prioritized
        scale/add attribute names.

        Parameters
        ----------
        attr : str | list | optional
            Name of scale factor or adder attribute. Can also
            be a prioritized list of attr names.
        default : float
            Default factor if attr is not found

        Returns
        -------
        factor : float
            Multiplicative or adder scale factor retrieved from dataset
            attributes.
        """

        factor = default

        if isinstance(attr, str):
            attr = [attr]

        for name in attr:
            if name in self.ds.attrs:
                factor = self.ds.attrs[name]
                break

        return factor

    @staticmethod
    def _check_slice(ds_slice):
        """
        Check ds_slice for lists, ensure lists are of the same len

        Parameters
        ----------
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from ds,
            each arg is for a sequential axis

        Returns
        -------
        list_len : int | None
            List lenght, None if none of the args are a list | ndarray
        multi_list : bool
            Flag if multiple list are provided in ds_slice
        """
        multi_list = False
        list_len = []
        for s in ds_slice:
            if isinstance(s, (list, np.ndarray)):
                list_len.append(len(s))

        if list_len:
            if len(list_len) > 1:
                multi_list = True

            list_len = list(set(list_len))
            if len(list_len) > 1:
                msg = ('shape mismatch: indexing arrays could not be '
                       'broadcast together with shapes {}'
                       .format(['({},)'.format(ln) for ln in list_len]))
                raise IndexError(msg)
            list_len = list_len[0]
        else:
            list_len = None

        return list_len, multi_list

    @staticmethod
    def _make_list_slices(ds_slice, list_len):
        """
        Duplicate slice arguements to enable zipping of list slices with
        non-list slices

        Parameters
        ----------
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from ds,
            each arg is for a sequential axis
        list_len : int
            List lenght

        Returns
        -------
        zip_slices : list
            List of slices to extract for each entry in list slice
        """
        zip_slices = []
        for s in ds_slice:
            if not isinstance(s, (list, np.ndarray)):
                zip_slices.append([s] * list_len)
            else:
                zip_slices.append(s)

        return zip_slices

    @staticmethod
    def _list_to_slice(ds_slice):
        """
        Check ds_slice to see if it is an int, slice, or list. Return
        pieces required for fancy indexing based on input type.

        Parameters
        ----------
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from ds,
            each arg is for a sequential axis

        Returns
        -------
        ds_slice : slice
            Slice that encompasses the entire range
        ds_idx : ndarray
            Adjusted list to extract points of interest from sliced array
        """
        ds_idx = None
        if isinstance(ds_slice, (list, np.ndarray)):
            in_slice = np.array(ds_slice)
            if np.issubdtype(in_slice.dtype, np.dtype(bool)):
                in_slice = np.where(in_slice)[0]

            s = in_slice.min()
            e = in_slice.max() + 1
            ds_slice = slice(s, e, None)
            ds_idx = in_slice - s
        elif isinstance(ds_slice, slice):
            ds_idx = slice(None)

        return ds_slice, ds_idx

    @staticmethod
    def _get_out_arr_slice(arr_slice, start):
        """
        Determine slice of pre-build output array that is being filled

        Parameters
        ----------
        arr_slice : tuple
            Tuple of (int, slice, list, ndarray) for section of output array
            being extracted
        start : int
            Start of slice, used for list gets

        Returns
        -------
        out_slice : tuple
            Slice arguments of portion of output array to insert arr_slice
            into
        stop : int
            Stop of slice, used for list gets, will be new start upon
            iteration
        """
        out_slice = ()
        int_slice = ()
        int_start = start
        int_stop = start
        stop = start
        for s in arr_slice:
            if isinstance(s, slice):
                out_slice += (slice(None), )
                int_slice += (slice(None), )
            elif isinstance(s, (int, np.integer)):
                if int_start == int_stop:
                    int_slice += (int_start, )
                    int_stop += 1
            elif isinstance(s, (list, tuple, np.ndarray)):
                list_len = len(s)
                if list_len == 1:
                    stop += 1
                    out_slice += ([start], )
                else:
                    stop += len(s)
                    out_slice += (slice(start, stop), )

        if not out_slice:
            out_slice += (start, )
            stop += 1
        elif all(s == slice(None) for s in out_slice):
            out_slice = int_slice
            stop = int_stop

        return out_slice, stop

    def _get_out_arr_shape(self, ds_slice):
        """
        Determine shape of output array

        Parameters
        ----------
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from ds,
            each arg is for a sequential axis

        Returns
        -------
        out_shape : tuple
            Shape of output array
        """
        ds_shape = self.shape
        out_shape = ()
        contains_list = False

        ds_slice += (slice(None), ) * (len(ds_shape) - len(ds_slice))
        for i, ax_slice in enumerate(ds_slice):
            if isinstance(ax_slice, slice):
                stop = ax_slice.stop
                if stop is None:
                    stop = ds_shape[i]

                out_shape += (len(range(*ax_slice.indices(stop))), )

            if isinstance(ax_slice, (list, tuple, np.ndarray)):
                if not contains_list:
                    out_shape += (len(ax_slice), )

                contains_list = True

        return out_shape

    def _extract_list_slice(self, ds_slice):
        """
        Optimize and extract list slice request along a single dimension. This
        function checks if sequential gid requests are more than one chunk size
        apart and then splits them into multiple separate requests (more
        efficient to do multipl reads than to read all gids in-between).

        Parameters
        ----------
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from ds,
            each arg is for a sequential axis

        Returns
        -------
        out : ndarray
            Extracted array of data from ds
        """
        out_slices = []
        chunks = self.chunks
        sort_idx = []
        list_len = None
        if chunks:
            for i, ax_slice in enumerate(ds_slice):
                c = chunks[i]
                if isinstance(ax_slice, (list, np.ndarray)):
                    if not isinstance(ax_slice, np.ndarray):
                        ax_slice = np.array(ax_slice)

                    idx = np.argsort(ax_slice)
                    sort_idx.append(np.argsort(idx))
                    ax_slice = ax_slice[idx]

                    # this checks if sequential gid requests are more than one
                    # chunk size apart and then splits them into multiple
                    # separate requests
                    diff = np.diff(ax_slice) > c
                    if np.any(diff):
                        pos = np.where(diff)[0] + 1
                        ax_slice = np.split(ax_slice, pos)
                        list_len = len(ax_slice)
                elif isinstance(ax_slice, slice):
                    sort_idx.append(slice(None))

                out_slices.append(ax_slice)
        else:
            out_slices = ds_slice

        if list_len is not None:
            out_shape = self._get_out_arr_shape(ds_slice)
            out_slices = self._make_list_slices(out_slices, list_len)

            out = np.zeros(out_shape, dtype=self.dtype)
            start = 0
            for s in zip(*out_slices):
                arr_slice, stop = self._get_out_arr_slice(s, start)
                out[arr_slice] = self._extract_ds_slice(s)
                start = stop

            out = out[tuple(sort_idx)]
        else:
            out = self._extract_ds_slice(ds_slice)

        return out

    def _extract_multi_list_slice(self, ds_slice, list_len):
        """
        Extract ds_slice that has multiple lists

        Parameters
        ----------
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from ds,
            each arg is for a sequential axis
        list_len : int
            List lenght

        Returns
        -------
        out : ndarray
            Extracted array of data from ds
        """
        zip_slices = self._make_list_slices(ds_slice, list_len)

        out_shape = self._get_out_arr_shape(ds_slice)

        out = np.zeros(out_shape, dtype=self.dtype)
        start = 0
        for s in zip(*zip_slices):
            arr_slice, stop = self._get_out_arr_slice(s, start)
            arr = self._extract_ds_slice(s)
            out[arr_slice] = arr

            start = stop

        return out

    def _extract_ds_slice(self, ds_slice):
        """
        Extact ds_slice from ds using slices where possible

        Parameters
        ----------
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from ds,
            each arg is for a sequential axis

        Returns
        -------
        out : ndarray
            Extracted array of data from ds
        """
        slices = ()
        idx_slice = ()
        for ax_slice in ds_slice:
            ax_slice, ax_idx = self._list_to_slice(ax_slice)
            slices += (ax_slice,)
            if ax_idx is not None:
                idx_slice += (ax_idx,)

        try:
            out = self.ds[slices]
        except Exception as e:
            msg = (f'Error retrieving data from "{self.ds}" for '
                   f'slice: "{slices}".')
            not_h5_ds = not isinstance(self.ds, h5py.Dataset)
            is_hsds_ds = 'h5pyd' in str(type(self.ds))
            error_type = isinstance(e, (OSError, IOError))
            if not_h5_ds and is_hsds_ds and error_type:
                msg += (' Detected OSError/IOError from h5pyd. '
                        'This is not a rex error and please do not submit '
                        'a bug report. this is likely due to HSDS server '
                        'limits, especially if you are using an NREL '
                        'developer API key. For more details, see: '
                        'https://nrel.github.io/rex/misc/examples.hsds.html')
            raise ResourceRuntimeError(msg) from e

        # check to see if idx_slice needs to be applied
        if any(s != slice(None) if isinstance(s, slice) else True
               for s in idx_slice):
            out = out[idx_slice]

        return out

    def _unscale_data(self, data):
        """
        Unscale dataset data

        Parameters
        ----------
        data : ndarray
            Native dataset array

        Returns
        -------
        data : ndarray
            Unscaled dataset array
        """
        data = data.astype('float32')

        if self.adder != 0:
            data *= self.scale_factor
            data += self.adder
        else:
            data /= self.scale_factor

        return data

    def _get_ds_slice(self, ds_slice):
        """
        Get ds_slice from ds as efficiently as possible, unscale if desired

        Parameters
        ----------
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from ds,
            each arg is for a sequential axis

        Returns
        -------
        out : ndarray
            Extracted array of data from ds
        """
        list_len, multi_list = self._check_slice(ds_slice)
        if list_len is not None:
            if multi_list:
                out = self._extract_multi_list_slice(ds_slice, list_len)
            else:
                out = self._extract_list_slice(ds_slice)
        else:
            out = self._extract_ds_slice(ds_slice)

        if self._unscale:
            out = self._unscale_data(out)

        return out

    @classmethod
    def extract(cls, ds, ds_slice, scale_attr='scale_factor',
                add_attr='add_offset', unscale=True):
        """
        Extract data from Resource Dataset

        Parameters
        ----------
        ds : h5py.dataset
            Open .h5 dataset instance to extract data from
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from ds,
            each arg is for a sequential axis
        scale_attr : str, optional
            Name of scale factor attribute, by default 'scale_factor'
        add_attr : str, optional
            Name of add offset attribute, by default 'add_offset'
        unscale : bool, optional
            Flag to unscale dataset data, by default True
        """
        dset = cls(ds, scale_attr=scale_attr, add_attr=add_attr,
                   unscale=unscale)

        return dset[ds_slice]


class BaseResource(BaseDatasetIterable):
    """
    Abstract Base class to handle resource .h5 files
    """

    SCALE_ATTR = 'scale_factor'
    ADD_ATTR = 'add_offset'
    UNIT_ATTR = 'units'

    def __init__(self, h5_file, mode='r', unscale=True, str_decode=True,
                 group=None, hsds=False, hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_file : str
            Path to .h5 resource file
        mode : str, optional
            Mode to instantiate h5py.File instance, by default 'r'
        unscale : bool, optional
            Boolean flag to automatically unscale variables on extraction,
            by default True
        str_decode : bool, optional
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read,
            by default True
        group : str, optional
            Group within .h5 resource file to open, by default None
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        self.h5_file = h5_file
        if hsds:
            if mode != 'r':
                raise OSError('Cannot write to files accessed vias HSDS!')

            import h5pyd
            if hsds_kwargs is None:
                hsds_kwargs = {}

            self._h5 = h5pyd.File(self.h5_file, mode='r', use_cache=False,
                                  **hsds_kwargs)
        else:
            try:
                self._h5 = h5py.File(self.h5_file, mode=mode)
            except Exception as e:
                msg = ('Could not open file in mode "{}": "{}"'
                       .format(mode, self.h5_file))
                raise OSError(msg) from e

        self._group = group
        self._unscale = unscale
        self._meta = None
        self._time_index = None
        self._lat_lon = None
        self._str_decode = str_decode
        self._attrs = None
        self._shapes = None
        self._chunks = None
        self._dtypes = None

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self.h5_file)

        return msg

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

        if type is not None:
            raise

    def __len__(self):
        return self.h5['time_index'].shape[0]

    def __getitem__(self, keys):
        ds, ds_slice = parse_keys(keys)
        _, ds_name = os.path.split(ds)

        if ds_name.startswith('time_index'):
            out = self._get_time_index(ds, ds_slice)

        elif ds_name.startswith('meta'):
            out = self._get_meta(ds, ds_slice)

        elif ds_name.startswith('coordinates'):
            out = self._get_coords(ds, ds_slice)

        elif 'SAM' in ds_name:
            site = ds_slice[0]
            if isinstance(site, (int, np.integer)):
                out = self.get_SAM_df(site)  # pylint: disable=E1111
            else:
                msg = "Can only extract SAM DataFrame for a single site"
                raise ResourceRuntimeError(msg)

        else:
            out = self._get_ds(ds, ds_slice)

        return out

    def __contains__(self, dset):
        return dset in self.datasets

    @classmethod
    def _get_datasets(cls, h5_obj, group=None):
        """
        Search h5 file instance for Datasets

        Parameters
        ----------
        h5_obj : h5py.File | h5py.Group
            Open h5py File or Group instance to search

        Returns
        -------
        dsets : list
            List of datasets in h5_obj
        """
        dsets = []
        for name in h5_obj:
            sub_obj = h5_obj[name]
            if isinstance(sub_obj, h5py.Group):
                dsets.extend(cls._get_datasets(sub_obj, group=name))
            else:
                dset_name = name
                if group is not None:
                    dset_name = "{}/{}".format(group, dset_name)

                dsets.append(dset_name)

        return dsets

    @property
    def h5(self):
        """
        Open h5py File instance. If _group is not None return open Group

        Returns
        -------
        h5 : h5py.File | h5py.Group
        """
        h5 = self._h5
        if self._group is not None:
            h5 = h5[self._group]

        return h5

    @property
    def datasets(self):
        """
        Datasets available

        Returns
        -------
        list
        """
        return self._get_datasets(self.h5)

    @property
    def dsets(self):
        """
        Datasets available

        Returns
        -------
        list
        """
        return self.datasets

    @property
    def resource_datasets(self):
        """
        Available resource datasets

        Returns
        -------
        list
        """
        dsets = [ds for ds in self.datasets
                 if ds not in ['meta', 'time_index', 'coordinates']]

        return dsets

    @property
    def res_dsets(self):
        """
        Available resource datasets

        Returns
        -------
        list
        """
        return self.resource_datasets

    @property
    def groups(self):
        """
        Groups available

        Returns
        -------
        groups : list
            List of groups
        """
        groups = []
        for name in self.h5:
            if isinstance(self.h5[name], h5py.Group):
                groups.append(name)

        return groups

    @property
    def shape(self):
        """
        Resource shape (timesteps, sites)
        shape = (len(time_index), len(meta))

        Returns
        -------
        shape : tuple
        """
        shape = (self.h5['time_index'].shape[0], self.h5['meta'].shape[0])

        return shape

    @property
    def meta(self):
        """
        Resource meta data DataFrame

        Returns
        -------
        meta : pandas.DataFrame
        """
        if self._meta is None:
            if 'meta' in self.h5:
                self._meta = self._get_meta('meta', slice(None))
            else:
                raise ResourceKeyError("'meta' is not a valid dataset")

        return self._meta

    @property
    def time_index(self):
        """
        Resource DatetimeIndex

        Returns
        -------
        time_index : pandas.DatetimeIndex
        """
        if self._time_index is None:
            if 'time_index' in self.h5:
                self._time_index = self._get_time_index('time_index',
                                                        slice(None))
            else:
                raise ResourceKeyError("'time_index' is not a valid dataset!")

        return self._time_index

    @property
    def coordinates(self):
        """
        Coordinates: (lat, lon) pairs

        Returns
        -------
        lat_lon : ndarray
        """
        return self.lat_lon

    @property
    def lat_lon(self):
        """
        Extract (latitude, longitude) pairs

        Returns
        -------
        lat_lon : ndarray
        """
        if self._lat_lon is None:
            if 'coordinates' in self:
                self._lat_lon = self._get_coords('coordinates', slice(None))
            else:
                lat_lon_cols = get_lat_lon_cols(self.meta)
                self._lat_lon = self.meta[lat_lon_cols].values

        return self._lat_lon

    @property
    def data_version(self):
        """
        Get the version attribute of the data. None if not available.

        Returns
        -------
        version : str | None
        """
        return self.global_attrs.get('version', None)

    @property
    def global_attrs(self):
        """
        Global (file) attributes

        Returns
        -------
        global_attrs : dict
        """
        return dict(self.h5.attrs)

    @property
    def attrs(self):
        """
        Dictionary of all dataset attributes

        Returns
        -------
        attrs : dict
        """
        if self._attrs is None:
            self._attrs = {}
            for dset in self.datasets:
                self._attrs[dset] = dict(self.h5[dset].attrs)

        return self._attrs

    @property
    def shapes(self):
        """
        Dictionary of all dataset shapes

        Returns
        -------
        shapes : dict
        """
        if self._shapes is None:
            self._shapes = {}
            for dset in self.datasets:
                self._shapes[dset] = self.h5[dset].shape

        return self._shapes

    @property
    def dtypes(self):
        """
        Dictionary of all dataset dtypes

        Returns
        -------
        dtypes : dict
        """
        if self._dtypes is None:
            self._dtypes = {}
            for dset in self.datasets:
                self._dtypes[dset] = self.h5[dset].dtype

        return self._dtypes

    @property
    def chunks(self):
        """
        Dictionary of all dataset chunk sizes

        Returns
        -------
        chunks : dict
        """
        if self._chunks is None:
            self._chunks = {}
            for dset in self.datasets:
                self._chunks[dset] = self._check_chunks(self.h5[dset].chunks)

        return self._chunks

    @property
    def adders(self):
        """
        Dictionary of all dataset add offset factors

        Returns
        -------
        adders : dict
        """
        return self._parse_attr_names(self.ADD_ATTR, 0)

    @property
    def scale_factors(self):
        """
        Dictionary of all dataset scale factors

        Returns
        -------
        scale_factors : dict
        """
        return self._parse_attr_names(self.SCALE_ATTR, 1)

    @property
    def units(self):
        """
        Dictionary of all dataset units

        Returns
        -------
        units : dict
        """
        return self._parse_attr_names(self.UNIT_ATTR, None)

    def _parse_attr_names(self, attr_names, default):
        """Retrieve an attribute from all dataset attributes.

        Parameters
        ----------
        attr_names : str | list
            Single or prioritized list of attribute names to retrieve, e.g.
            "scale_factor" or ["scale_factor", "psm_scale_factor"]
        default : None | int | float
            Default value if attr_names not found in any given dataset.

        Returns
        -------
        out : dict
            Dictionary mapping datasets (keys) to attribute values (values),
            e.g. if attr_names="scale_factor", out would be:
            {"windspeed_10m": 10, "misc": 1}
        """

        if isinstance(attr_names, str):
            attr_names = [attr_names]

        out = {}
        for dset, attrs in self.attrs.items():
            out[dset] = default
            for name in attr_names:
                if name in attrs:
                    out[dset] = attrs[name]
                    break

        return out

    @staticmethod
    def _check_chunks(chunks):
        """
        Check to see if chunks is an HSDS dictionary, if so convert to a tuple

        Parameters
        ----------
        chunks : tuple | dict | None
            tuple of chunk size, None, or HSDS chunk dictionary

        Returns
        -------
        chunks : tuple
            Tuple of chunk size along all axes
        """
        if isinstance(chunks, dict):
            chunks = tuple(chunks.get('dims', None))

        return chunks

    @staticmethod
    def df_str_decode(df):
        """Decode a dataframe with byte string columns into ordinary str cols.

        Parameters
        ----------
        df : pd.DataFrame
            Dataframe with some columns being byte strings.

        Returns
        -------
        df : pd.DataFrame
            DataFrame with str columns instead of byte str columns.
        """
        for col in df:
            if (np.issubdtype(df[col].dtype, np.object_)
                    and isinstance(df[col].values[0], bytes)):
                df[col] = df[col].copy().str.decode('utf-8', 'ignore')

        return df

    def open_dataset(self, ds_name):
        """
        Open resource dataset

        Parameters
        ----------
        ds_name : str
            Dataset name to open

        Returns
        -------
        ds : ResourceDataset
            Resource for open resource dataset
        """
        if ds_name not in self.datasets:
            raise ResourceKeyError('{} not in {}'
                                   .format(ds_name, self.datasets))

        ds = ResourceDataset(self.h5[ds_name], scale_attr=self.SCALE_ATTR,
                             add_attr=self.ADD_ATTR, unscale=self._unscale)

        return ds

    def get_attrs(self, dset=None):
        """
        Get h5 attributes either from file or dataset

        Parameters
        ----------
        dset : str
            Dataset to get attributes for, if None get file (global) attributes

        Returns
        -------
        attrs : dict
            Dataset or file attributes
        """
        if dset is None:
            attrs = dict(self.h5.attrs)
        else:
            attrs = dict(self.h5[dset].attrs)

        return attrs

    def get_dset_properties(self, dset):
        """
        Get dataset properties (shape, dtype, chunks)

        Parameters
        ----------
        dset : str
            Dataset to get scale factor for

        Returns
        -------
        shape : tuple
            Dataset array shape
        dtype : str
            Dataset array dtype
        chunks : tuple
            Dataset chunk size
        """
        ds = self.h5[dset]
        shape, dtype, chunks = ds.shape, ds.dtype, ds.chunks

        return shape, dtype, self._check_chunks(chunks)

    def get_scale_factor(self, dset):
        """
        Get dataset scale factor

        Parameters
        ----------
        dset : str
            Dataset to get scale factor for

        Returns
        -------
        float
            Dataset scale factor, used to unscale int values to floats
        """
        return self.scale_factors[dset]

    # pylint: disable=redefined-argument-from-local
    def get_units(self, dset):
        """
        Get dataset units

        Parameters
        ----------
        dset : str
            Dataset to get units for

        Returns
        -------
        str
            Dataset units, None if not defined
        """
        if dset not in self:
            name = dset.split('_')[0]
            for dset in self.resource_datasets:
                if dset.startswith(name):
                    break

        return self.units[dset]

    def get_meta_arr(self, rec_name, rows=slice(None)):
        """Get a meta array by name (faster than DataFrame extraction).

        Parameters
        ----------
        rec_name : str
            Named record from the meta data to retrieve.
        rows : slice
            Rows of the record to extract.

        Returns
        -------
        meta_arr : np.ndarray
            Extracted array from the meta data record name.
        """
        if 'meta' in self.h5:
            meta_arr = self.h5['meta'][rec_name, rows]
            if self._str_decode and np.issubdtype(meta_arr.dtype, np.bytes_):
                meta_arr = np.char.decode(meta_arr, encoding='utf-8')
        else:
            raise ResourceKeyError("'meta' is not a valid dataset")

        return meta_arr

    def _get_time_index(self, ds_name, ds_slice):
        """
        Extract and convert time_index to pandas Datetime Index

        Parameters
        ----------
        ds_name : str
            Dataset to extract time_index from
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from
            time_index

        Returns
        -------
        time_index : pandas.DatetimeIndex
            Vector of datetime stamps
        """
        ds_slice = parse_slice(ds_slice)
        time_index = self.h5[ds_name]
        time_index = ResourceDataset.extract(time_index, ds_slice[0],
                                             unscale=False)
        try:
            datetime_index = pd.to_datetime(time_index.astype(str))
        except (pd.errors.OutOfBoundsDatetime, dateutil.parser.ParserError):
            return time_index

        return check_tz(datetime_index)

    def _get_meta(self, ds_name, ds_slice):
        """
        Extract and convert meta to a pandas DataFrame

        Parameters
        ----------
        ds_name : str
            Dataset to extract meta from
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray, str) of what sites and columns
            to extract from meta

        Returns
        -------
        meta : pandas.Dataframe
            Dataframe of location meta data
        """
        ds_slice = parse_slice(ds_slice)
        sites = ds_slice[0]
        if isinstance(sites, (int, np.integer)):
            sites = slice(sites, sites + 1)

        meta = self.h5[ds_name]
        meta = ResourceDataset.extract(meta, sites, unscale=False)

        if isinstance(sites, slice):
            stop = sites.stop
            if stop is None:
                stop = len(meta)

            sites = list(range(*sites.indices(stop)))

        meta = pd.DataFrame(meta, index=sites)
        if 'gid' not in meta:
            meta.index.name = 'gid'

        if self._str_decode:
            meta = self.df_str_decode(meta)

        if len(ds_slice) == 2:
            meta = meta[ds_slice[1]]

        return meta

    def _get_coords(self, ds_name, ds_slice):
        """
        Extract coordinates (lat, lon) pairs

        Parameters
        ----------
        ds_name : str
            Dataset to extract coordinates from
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from
            coordinates, each arg is for a sequential axis

        Returns
        -------
        coords : ndarray
            Array of (lat, lon) pairs for each site in meta
        """
        ds_slice = parse_slice(ds_slice)
        coords = self.h5[ds_name]
        coords = ResourceDataset.extract(coords, ds_slice[0],
                                         unscale=False)
        return coords

    # pylint: disable=unused-argument
    def get_SAM_df(self, site):
        """
        Placeholder for get_SAM_df method that it resource specific

        Parameters
        ----------
        site : int
            Site to extract SAM DataFrame for
        """
        msg = ('Method to retrieve SAM dataframe not implemented for vanilla '
               'Resource handler. Use an NSRDB or WTK handler instead.')
        raise NotImplementedError(msg)

    def _get_ds(self, ds_name, ds_slice):
        """
        Extract data from given dataset

        Parameters
        ----------
        ds_name : str
            Variable dataset to be extracted
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from ds,
            each arg is for a sequential axis

        Returns
        -------
        out : ndarray
            ndarray of variable timeseries data
            If unscale, returned in native units else in scaled units
        """
        if ds_name not in self.datasets:
            raise ResourceKeyError('{} not in {}'
                                   .format(ds_name, self.datasets))

        ds = self.h5[ds_name]
        ds_slice = parse_slice(ds_slice)
        if len(ds_slice) > len(ds.shape):
            return self._get_ds_with_repeated_values(ds, ds_name, ds_slice)
        return ResourceDataset.extract(ds, ds_slice,
                                       scale_attr=self.SCALE_ATTR,
                                       add_attr=self.ADD_ATTR,
                                       unscale=self._unscale)

    def _get_ds_with_repeated_values(self, ds, ds_name, ds_slice):
        """
        Extract 1D data using 2D slice by repeating the 1D data along
        the spatial or temporal dimension

        Parameters
        ----------
        ds : h5py.dataset
            Open .h5 dataset instance to extract data from
        ds_name : str
            Variable dataset to be extracted
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from ds,
            each arg is for a sequential axis

        Returns
        -------
        out : ndarray
            ndarray of variable timeseries data
            If unscale, returned in native units else in scaled units
        """
        ti_shape = self.shapes.get('time_index')
        meta_shape = self.shapes.get('meta')
        if ti_shape == meta_shape:
            msg = ("Attempting to use a 2D slice on a 1D dataset when the "
                   "meta and time index have the same shape - unable to "
                   "disambiguate the slice dimensions. Please either update "
                   "the length of the meta and/or index, set the shape of "
                   "{!r} to be 2-dimensional (current shape: {!r}), or use a "
                   "1-dimensional slice.".format(ds_name, ds.shape))
            raise ResourceRuntimeError(msg)

        if ds.shape == ti_shape:
            return self._get_ds_with_spatial_repeat(ds, ds_name, ds_slice)
        if ds.shape == meta_shape:
            return self._get_ds_with_temporal_repeat(ds, ds_name, ds_slice)

        msg = ("Attempting to use a 2D slice on a 1D dataset ({0!r}) when "
               "the shape of the dataset {1!r} does not match the shape "
               "of the meta {2!r})or the time index {3!r}. Please either "
               "update the length of ({0!r}) to match either the meta or "
               "index, or use a 1-dimensional slice."
               .format(ds_name, ds.shape, meta_shape, ti_shape))
        raise ResourceRuntimeError(msg)

    def _get_ds_with_spatial_repeat(self, ds, ds_name, ds_slice):
        """
        Extract 1D data using 2D slice by repeating the 1D data along
        the spatial dimension

        Parameters
        ----------
        ds : h5py.dataset
            Open .h5 dataset instance to extract data from
        ds_name : str
            Variable dataset to be extracted
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from ds,
            each arg is for a sequential axis

        Returns
        -------
        out : ndarray
            ndarray of variable timeseries data
            If unscale, returned in native units else in scaled units
        """
        msg = ("Attempting to use a 2D slice on a 1D dataset ({!r}). 1D "
               "dataset values will be repeated along the spatial dimension"
               .format(ds_name))
        warn(msg)

        out = ResourceDataset.extract(ds, ds_slice[:1],
                                      scale_attr=self.SCALE_ATTR,
                                      add_attr=self.ADD_ATTR,
                                      unscale=self._unscale)
        if not isinstance(out, np.ndarray):
            out *= np.ones(self.shapes['meta'], dtype=np.float32)
            out = out[ds_slice[1]]
        else:
            out = np.repeat(out[:, None], self.shapes['meta'][0], axis=1)
            out = out[:, ds_slice[1]]

        return out.astype(np.float32)

    def _get_ds_with_temporal_repeat(self, ds, ds_name, ds_slice):
        """
        Extract 1D data using 2D slice by repeating the 1D data along
        the temporal dimension

        Parameters
        ----------
        ds : h5py.dataset
            Open .h5 dataset instance to extract data from
        ds_name : str
            Variable dataset to be extracted
        ds_slice : tuple
            Tuple of (int, slice, list, ndarray) of what to extract from ds,
            each arg is for a sequential axis

        Returns
        -------
        out : ndarray
            ndarray of variable timeseries data
            If unscale, returned in native units else in scaled units
        """
        msg = ("Attempting to use a 2D slice on a 1D dataset ({!r}). 1D "
               "dataset values will be repeated along the temporal dimension"
               .format(ds_name))
        warn(msg)

        out = ResourceDataset.extract(ds, ds_slice[1:],
                                      scale_attr=self.SCALE_ATTR,
                                      add_attr=self.ADD_ATTR,
                                      unscale=self._unscale)

        if not isinstance(out, np.ndarray):
            out *= np.ones(self.shapes['time_index'], dtype=np.float32)
        else:
            out = np.ones((self.shapes['time_index'][0], len(out))) * out

        return out[ds_slice[0]].astype(np.float32)

    def close(self):
        """
        Close h5 instance
        """
        self._h5.close()

    @staticmethod
    def _preload_SAM(res, sites, tech, time_index_step=None, means=False):
        """
        Placeholder method to pre-load project_points for SAM

        Parameters
        ----------
        res : rex.Resource
            rex Resource handler or similar (NSRDB, WindResource,
            MultiFileResource, etc...)
        sites : list
            List of sites to be provided to SAM
            (sites is synonymous with gids aka spatial indices)
        tech : str
            Technology to be run by SAM
        time_index_step: int, optional
            Step size for time_index, used to reduce temporal resolution,
            by default None
        means : bool, optional
            Boolean flag to compute mean resource when res_array is set,
            by default False
        """
        time_slice = slice(None, None, time_index_step)
        SAM_res = SAMResource(sites, tech, res['time_index', time_slice],
                              means=means)
        sites = SAM_res.sites_slice
        SAM_res['meta'] = res['meta', sites]

        SAM_res.load_rex_resource(res, SAM_res.var_list, time_slice, sites)

        return SAM_res

    @classmethod
    def preload_SAM(cls, h5_file, sites, tech, unscale=True, str_decode=True,
                    group=None, hsds=False, hsds_kwargs=None,
                    time_index_step=None, means=False):
        """
        Pre-load project_points for SAM

        Parameters
        ----------
        h5_file : str
            h5_file to extract resource from
        sites : list
            List of sites to be provided to SAM
            (sites is synonymous with gids aka spatial indices)
        tech : str
            Technology to be run by SAM
        unscale : bool
            Boolean flag to automatically unscale variables on extraction
        str_decode : bool
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read.
        group : str
            Group within .h5 resource file to open
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        time_index_step: int, optional
            Step size for time_index, used to reduce temporal resolution,
            by default None
        means : bool, optional
            Boolean flag to compute mean resource when res_array is set,
            by default False

        Returns
        -------
        SAM_res : SAMResource
            Instance of SAMResource pre-loaded with Solar resource for sites
            in project_points
        """
        kwargs = {"unscale": unscale, "hsds": hsds, 'hsds_kwargs': hsds_kwargs,
                  "str_decode": str_decode, "group": group}
        with cls(h5_file, **kwargs) as res:
            SAM_res = res._preload_SAM(res, sites, tech,
                                       time_index_step=time_index_step,
                                       means=means)

        return SAM_res


class Resource(BaseResource):
    """
    Base class to handle resource .h5 files

    Examples
    --------
    Extracting the resource's Datetime Index

    >>> file = '$TESTDATADIR/nsrdb/ri_100_nsrdb_2012.h5'
    >>> with Resource(file) as res:
    >>>     ti = res.time_index
    >>>
    >>> ti
    DatetimeIndex(['2012-01-01 00:00:00', '2012-01-01 00:30:00',
                   '2012-01-01 01:00:00', '2012-01-01 01:30:00',
                   '2012-01-01 02:00:00', '2012-01-01 02:30:00',
                   '2012-01-01 03:00:00', '2012-01-01 03:30:00',
                   '2012-01-01 04:00:00', '2012-01-01 04:30:00',
                   ...
                   '2012-12-31 19:00:00', '2012-12-31 19:30:00',
                   '2012-12-31 20:00:00', '2012-12-31 20:30:00',
                   '2012-12-31 21:00:00', '2012-12-31 21:30:00',
                   '2012-12-31 22:00:00', '2012-12-31 22:30:00',
                   '2012-12-31 23:00:00', '2012-12-31 23:30:00'],
                  dtype='datetime64[ns]', length=17568, freq=None)

    Efficient slicing of the Datetime Index

    >>> with Resource(file) as res:
    >>>     ti = res['time_index', 1]
    >>>
    >>> ti
    2012-01-01 00:30:00

    >>> with Resource(file) as res:
    >>>     ti = res['time_index', :10]
    >>>
    >>> ti
    DatetimeIndex(['2012-01-01 00:00:00', '2012-01-01 00:30:00',
                   '2012-01-01 01:00:00', '2012-01-01 01:30:00',
                   '2012-01-01 02:00:00', '2012-01-01 02:30:00',
                   '2012-01-01 03:00:00', '2012-01-01 03:30:00',
                   '2012-01-01 04:00:00', '2012-01-01 04:30:00'],
                  dtype='datetime64[ns]', freq=None)

    >>> with Resource(file) as res:
    >>>     ti = res['time_index', [1, 2, 4, 8, 9]
    >>>
    >>> ti
    DatetimeIndex(['2012-01-01 00:30:00', '2012-01-01 01:00:00',
                   '2012-01-01 02:00:00', '2012-01-01 04:00:00',
                   '2012-01-01 04:30:00'],
                  dtype='datetime64[ns]', freq=None)

    Extracting resource's site metadata

    >>> with Resource(file) as res:
    >>>     meta = res.meta
    >>>
    >>> meta
            latitude  longitude   elevation  timezone    country ...
    0      41.29     -71.86    0.000000        -5           None ...
    1      41.29     -71.82    0.000000        -5           None ...
    2      41.25     -71.82    0.000000        -5           None ...
    3      41.33     -71.82   15.263158        -5  United States ...
    4      41.37     -71.82   25.360000        -5  United States ...
    ..       ...        ...         ...       ...            ... ...
    95     41.25     -71.66    0.000000        -5           None ...
    96     41.89     -71.66  153.720000        -5  United States ...
    97     41.45     -71.66   35.440000        -5  United States ...
    98     41.61     -71.66  140.200000        -5  United States ...
    99     41.41     -71.66   35.160000        -5  United States ...
    [100 rows x 10 columns]

    Efficient slicing of the metadata

    >>> with Resource(file) as res:
    >>>     meta = res['meta', 1]
    >>>
    >>> meta
       latitude  longitude  elevation  timezone country state county urban ...
    1     41.29     -71.82        0.0        -5    None  None   None  None ...

    >>> with Resource(file) as res:
    >>>     meta = res['meta', :5]
    >>>
    >>> meta
       latitude  longitude  elevation  timezone        country ...
    0     41.29     -71.86   0.000000        -5           None ...
    1     41.29     -71.82   0.000000        -5           None ...
    2     41.25     -71.82   0.000000        -5           None ...
    3     41.33     -71.82  15.263158        -5  United States ...
    4     41.37     -71.82  25.360000        -5  United States ...

    >>> with Resource(file) as res:
    >>>     tz = res['meta', :, 'timezone']
    >>>
    >>> tz
    0    -5
    1    -5
    2    -5
    3    -5
    4    -5
         ..
    95   -5
    96   -5
    97   -5
    98   -5
    99   -5
    Name: timezone, Length: 100, dtype: int64

    >>> with Resource(file) as res:
    >>>     lat_lon = res['meta', :, ['latitude', 'longitude']]
    >>>
    >>> lat_lon
        latitude  longitude
    0      41.29     -71.86
    1      41.29     -71.82
    2      41.25     -71.82
    3      41.33     -71.82
    4      41.37     -71.82
    ..       ...        ...
    95     41.25     -71.66
    96     41.89     -71.66
    97     41.45     -71.66
    98     41.61     -71.66
    99     41.41     -71.66
    [100 rows x 2 columns]

    Extracting resource variables (datasets)

    >>> with Resource(file) as res:
    >>>     wspd = res['wind_speed']
    >>>
    >>> wspd
    [[12. 12. 12. ... 12. 12. 12.]
     [12. 12. 12. ... 12. 12. 12.]
     [12. 12. 12. ... 12. 12. 12.]
     ...
     [14. 14. 14. ... 14. 14. 14.]
     [15. 15. 15. ... 15. 15. 15.]
     [15. 15. 15. ... 15. 15. 15.]]

    Efficient slicing of variables

    >>> with Resource(file) as res:
    >>>     wspd = res['wind_speed', :2]
    >>>
    >>> wspd
    [[12. 12. 12. 12. 12. 12. 53. 53. 53. 53. 53. 12. 53.  1.  1. 12. 12. 12.
       1.  1. 12. 53. 53. 53. 12. 12. 12. 12. 12.  1. 12. 12.  1. 12. 12. 53.
      12. 53.  1. 12.  1. 53. 53. 12. 12. 12. 12.  1.  1.  1. 12. 12.  1.  1.
      12. 12. 53. 53. 53. 12. 12. 53. 53. 12. 12. 12. 12. 12. 12.  1. 53.  1.
      53. 12. 12. 53. 53.  1.  1.  1. 53. 12.  1.  1. 53. 53. 53. 12. 12. 12.
      12. 12. 12. 12.  1. 12.  1. 12. 12. 12.]
     [12. 12. 12. 12. 12. 12. 53. 53. 53. 53. 53. 12. 53.  1.  1. 12. 12. 12.
       1.  1. 12. 53. 53. 53. 12. 12. 12. 12. 12.  1. 12. 12.  1. 12. 12. 53.
      12. 53.  1. 12.  1. 53. 53. 12. 12. 12. 12.  1.  1.  1. 12. 12.  1.  1.
      12. 12. 53. 53. 53. 12. 12. 53. 53. 12. 12. 12. 12. 12. 12.  1. 53.  1.
      53. 12. 12. 53. 53.  1.  1.  1. 53. 12.  1.  1. 53. 53. 53. 12. 12. 12.
      12. 12. 12. 12.  1. 12.  1. 12. 12. 12.]]

    >>> with Resource(file) as res:
    >>>     wspd = res['wind_speed', :, [2, 3]]
    >>>
    >>> wspd
    [[12. 12.]
     [12. 12.]
     [12. 12.]
     ...
     [14. 14.]
     [15. 15.]
     [15. 15.]]
    """

    SCALE_ATTR = 'scale_factor'
    ADD_ATTR = 'add_offset'
    UNIT_ATTR = 'units'

    def __init__(self, h5_file, unscale=True, str_decode=True, group=None,
                 hsds=False, hsds_kwargs=None):
        """
        Parameters
        ----------
        h5_file : str
            Path to .h5 resource file
        unscale : bool, optional
            Boolean flag to automatically unscale variables on extraction,
            by default True
        str_decode : bool, optional
            Boolean flag to decode the bytestring meta data into normal
            strings. Setting this to False will speed up the meta data read,
            by default True
        group : str, optional
            Group within .h5 resource file to open, by default None
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        hsds_kwargs : dict, optional
            Dictionary of optional kwargs for h5pyd, e.g., bucket, username,
            password, by default None
        """
        super().__init__(h5_file, unscale=unscale, str_decode=str_decode,
                         group=group, mode='r', hsds=hsds,
                         hsds_kwargs=hsds_kwargs)
